# -*- coding: utf-8 -*-
"""Interface to the file repository of a node instance."""
import contextlib
import io
import pathlib
import tempfile
from typing import BinaryIO, Dict, Iterable, Iterator, List, Tuple, Union

from aiida.common import exceptions
from aiida.repository import File, Repository
from aiida.repository.backend import SandboxRepositoryBackend

__all__ = ('NodeRepositoryMixin',)

FilePath = Union[str, pathlib.PurePosixPath]


class NodeRepositoryMixin:
    """Interface to the file repository of a node instance.

    This is the compatibility layer between the `Node` class and the `Repository` class. The repository in principle has
    no concept of immutability, so it is implemented here. Any mutating operations will raise a `ModificationNotAllowed`
    exception if the node is stored. Otherwise the operation is just forwarded to the repository instance.

    The repository instance keeps an internal mapping of the file hierarchy that it maintains, starting from an empty
    hierarchy if the instance was constructed normally, or from a specific hierarchy if reconstructred through the
    ``Repository.from_serialized`` classmethod. This is only the case for stored nodes, because unstored nodes do not
    have any files yet when they are constructed. Once the node get's stored, the repository is asked to serialize its
    metadata contents which is then stored in the ``repository_metadata`` attribute of the node in the database. This
    layer explicitly does not update the metadata of the node on a mutation action. The reason is that for stored nodes
    these actions are anyway forbidden and for unstored nodes, the final metadata will be stored in one go, once the
    node is stored, so there is no need to keep updating the node metadata intermediately. Note that this does mean that
    ``repository_metadata`` does not give accurate information as long as the node is not yet stored.
    """

    _repository_instance = None

    def _update_repository_metadata(self):
        """Refresh the repository metadata of the node if it is stored and the decorated method returns successfully."""
        if self.is_stored:
            self.repository_metadata = self._repository.serialize()

    @property
    def _repository(self) -> Repository:
        """Return the repository instance, lazily constructing it if necessary.

        .. note:: this property is protected because a node's repository should not be accessed outside of its scope.

        :return: the file repository instance.
        """
        if self._repository_instance is None:
            if self.is_stored:
                backend = self.backend.get_repository()
                serialized = self.repository_metadata
                self._repository_instance = Repository.from_serialized(backend=backend, serialized=serialized)
            else:
                self._repository_instance = Repository(backend=SandboxRepositoryBackend())

        return self._repository_instance

    @_repository.setter
    def _repository(self, repository: Repository) -> None:
        """Set a new repository instance, deleting the current reference if it has been initialized.

        :param repository: the new repository instance to set.
        """
        if self._repository_instance is not None:
            del self._repository_instance

        self._repository_instance = repository

    def repository_serialize(self) -> Dict:
        """Serialize the metadata of the repository content into a JSON-serializable format.

        :return: dictionary with the content metadata.
        """
        return self._repository.serialize()

    def check_mutability(self):
        """Check if the node is mutable.

        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the node is stored and therefore the repository is immutable.')

    def list_objects(self, path: str = None) -> List[File]:
        """Return a list of the objects contained in this repository sorted by name, optionally in given sub directory.

        :param path: the relative path where to store the object in the repository.
        :return: a list of `File` named tuples representing the objects present in directory with the given key.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if no object exists for the given path.
        :raises NotADirectoryError: if the object at the given path is not a directory.
        """
        return self._repository.list_objects(path)

    def list_object_names(self, path: str = None) -> List[str]:
        """Return a sorted list of the object names contained in this repository, optionally in the given sub directory.

        :param path: the relative path where to store the object in the repository.
        :return: a list of `File` named tuples representing the objects present in directory with the given key.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if no object exists for the given path.
        :raises NotADirectoryError: if the object at the given path is not a directory.
        """
        return self._repository.list_object_names(path)

    @contextlib.contextmanager
    def open(self, path: str, mode='r') -> Iterator[BinaryIO]:
        """Open a file handle to an object stored under the given key.

        .. note:: this should only be used to open a handle to read an existing file. To write a new file use the method
            ``put_object_from_filelike`` instead.

        :param path: the relative path of the object within the repository.
        :return: yield a byte stream object.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if the file does not exist.
        :raises IsADirectoryError: if the object is a directory and not a file.
        :raises OSError: if the file could not be opened.
        """
        if mode not in ['r', 'rb']:
            raise ValueError(f'the mode {mode} is not supported.')

        with self._repository.open(path) as handle:
            if 'b' not in mode:
                yield io.StringIO(handle.read().decode('utf-8'))
            else:
                yield handle

    def get_object(self, path: FilePath = None) -> File:
        """Return the object at the given path.

        :param path: the relative path where to store the object in the repository.
        :return: the `File` representing the object located at the given relative path.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if no object exists for the given path.
        """
        return self._repository.get_object(path)

    def get_object_content(self, path: str, mode='r') -> Union[str, bytes]:
        """Return the content of a object identified by key.

        :param key: fully qualified identifier for the object within the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if the file does not exist.
        :raises IsADirectoryError: if the object is a directory and not a file.
        :raises OSError: if the file could not be opened.
        """
        if mode not in ['r', 'rb']:
            raise ValueError(f'the mode {mode} is not supported.')

        if 'b' not in mode:
            return self._repository.get_object_content(path).decode('utf-8')

        return self._repository.get_object_content(path)

    def put_object_from_filelike(self, handle: io.BufferedReader, path: str):
        """Store the byte contents of a file in the repository.

        :param handle: filelike object with the byte content to be stored.
        :param path: the relative path where to store the object in the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        self.check_mutability()

        if isinstance(handle, io.StringIO):
            handle = io.BytesIO(handle.read().encode('utf-8'))

        if isinstance(handle, tempfile._TemporaryFileWrapper):  # pylint: disable=protected-access
            if 'b' in handle.file.mode:
                handle = io.BytesIO(handle.read())
            else:
                handle = io.BytesIO(handle.read().encode('utf-8'))

        self._repository.put_object_from_filelike(handle, path)
        self._update_repository_metadata()

    def put_object_from_file(self, filepath: str, path: str):
        """Store a new object under `path` with contents of the file located at `filepath` on the local file system.

        :param filepath: absolute path of file whose contents to copy to the repository
        :param path: the relative path where to store the object in the repository.
        :raises TypeError: if the path is not a string and relative path, or the handle is not a byte stream.
        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        self.check_mutability()
        self._repository.put_object_from_file(filepath, path)
        self._update_repository_metadata()

    def put_object_from_tree(self, filepath: str, path: str = None):
        """Store the entire contents of `filepath` on the local file system in the repository with under given `path`.

        :param filepath: absolute path of the directory whose contents to copy to the repository.
        :param path: the relative path where to store the objects in the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        self.check_mutability()
        self._repository.put_object_from_tree(filepath, path)
        self._update_repository_metadata()

    def walk(self, path: FilePath = None) -> Iterable[Tuple[pathlib.PurePosixPath, List[str], List[str]]]:
        """Walk over the directories and files contained within this repository.

        .. note:: the order of the dirname and filename lists that are returned is not necessarily sorted. This is in
            line with the ``os.walk`` implementation where the order depends on the underlying file system used.

        :param path: the relative path of the directory within the repository whose contents to walk.
        :return: tuples of root, dirnames and filenames just like ``os.walk``, with the exception that the root path is
            always relative with respect to the repository root, instead of an absolute path and it is an instance of
            ``pathlib.PurePosixPath`` instead of a normal string
        """
        yield from self._repository.walk(path)

    def glob(self) -> Iterable[pathlib.PurePosixPath]:
        """Yield a recursive list of all paths (files and directories)."""
        for dirpath, dirnames, filenames in self.walk():
            for dirname in dirnames:
                yield dirpath / dirname
            for filename in filenames:
                yield dirpath / filename

    def copy_tree(self, target: Union[str, pathlib.Path], path: FilePath = None) -> None:
        """Copy the contents of the entire node repository to another location on the local file system.

        :param target: absolute path of the directory where to copy the contents to.
        :param path: optional relative path whose contents to copy.
        """
        self._repository.copy_tree(target, path)

    def delete_object(self, path: str):
        """Delete the object from the repository.

        :param key: fully qualified identifier for the object within the repository.
        :raises TypeError: if the path is not a string and relative path.
        :raises FileNotFoundError: if the file does not exist.
        :raises IsADirectoryError: if the object is a directory and not a file.
        :raises OSError: if the file could not be deleted.
        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        self.check_mutability()
        self._repository.delete_object(path)
        self._update_repository_metadata()

    def erase(self):
        """Delete all objects from the repository.

        :raises `~aiida.common.exceptions.ModificationNotAllowed`: when the node is stored and therefore immutable.
        """
        self.check_mutability()
        self._repository.erase()
        self._update_repository_metadata()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Defines reStructuredText directives to simplify documenting AiiDA and its plugins."""


def setup(app):
    """Setup function to add the extension classes / nodes to Sphinx."""
    import aiida

    from . import calcjob, process, workchain

    app.setup_extension('sphinxcontrib.details.directive')
    process.setup_extension(app)
    workchain.setup_extension(app)
    calcjob.setup_extension(app)

    return {'version': aiida.__version__, 'parallel_read_safe': True}



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module
"""Base SQLAlchemy models."""
from sqlalchemy import MetaData
from sqlalchemy.orm import declarative_base


class Model:
    """Base ORM model."""


# see https://alembic.sqlalchemy.org/en/latest/naming.html
naming_convention = (
    ('pk', '%(table_name)s_pkey'),  # this is identical to the default PSQL convention
    ('ix', 'ix_%(table_name)s_%(column_0_N_label)s'),
    # note, indexes using varchar_pattern_ops should be named: 'ix_pat_%(table_name)s_%(column_0_N_label)s'
    ('uq', 'uq_%(table_name)s_%(column_0_N_name)s'),
    ('ck', 'ck_%(table_name)s_%(constraint_name)s'),
    ('fk', 'fk_%(table_name)s_%(column_0_N_name)s_%(referred_table_name)s'),
    # note, ideally we may also append with '_%(referred_column_0_N_name)s', but this causes ORM construction errors:
    # https://github.com/sqlalchemy/sqlalchemy/issues/5350
)

Base = declarative_base(cls=Model, name='Model', metadata=MetaData(naming_convention=dict(naming_convention)))  # pylint: disable=invalid-name


def get_orm_metadata() -> MetaData:
    """Return the populated metadata object."""
    # we must load all models, to populate the ORM metadata
    from aiida.storage.psql_dos.models import (  # pylint: disable=unused-import
        authinfo,
        comment,
        computer,
        group,
        log,
        node,
        settings,
        user,
    )
    return Base.metadata



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility functions to operate on datetime objects."""
from datetime import datetime

from dateutil import parser


def get_current_timezone():
    """Return the current timezone.

    :return: current timezone
    """
    from tzlocal import get_localzone
    local = get_localzone()

    if local.zone == 'local':
        raise ValueError(
            "Unable to detect name of local time zone. Please set 'TZ' environment variable, e.g. to 'Europe/Zurich'"
        )
    return local


def now():
    """Return the datetime object of the current time.

    :return: datetime object represeting current time
    """
    import pytz

    from aiida.manage.configuration import settings

    if getattr(settings, 'USE_TZ', None):
        return datetime.utcnow().replace(tzinfo=pytz.utc)

    return datetime.now()


def is_naive(value):
    """Return whether the given datetime object is timezone naive

    :return: boolean, True if naive, False otherwise
    """
    return value.utcoffset() is None


def is_aware(value):
    """Return whether the given datetime object is timezone aware

    :return: boolean, True if aware, False otherwise
    """
    return value.utcoffset() is not None


def make_aware(value, timezone=None, is_dst=None):
    """Make the given datetime object timezone aware.

    :param value: datetime object to make aware
    :param timezone:
    :param is_dst:
    :return:
    """
    if timezone is None:
        timezone = get_current_timezone()
    if hasattr(timezone, 'localize'):
        return timezone.localize(value, is_dst=is_dst)

    if is_aware(value):
        raise ValueError(f'make_aware expects a naive datetime, got {value}')

    # This may be wrong around DST changes!
    # See http://pytz.sourceforge.net/#localized-times-and-date-arithmetic
    return timezone.localize(value)


def localtime(value, timezone=None):
    """Converts an aware datetime.datetime to local time.

    Local time is defined by the current time zone, unless another time zone is specified.
    """
    if timezone is None:
        timezone = get_current_timezone()

    # If `value` is naive, astimezone() will raise a ValueError, so we don't need to perform a redundant check.
    value = value.astimezone(timezone)
    if hasattr(timezone, 'normalize'):
        # This method is available for pytz time zones.
        value = timezone.normalize(value)

    return value


def delta(from_time, to_time=None):
    """Return the datetime object representing the different between two datetime objects.

    :param from_time: starting datetime object
    :param to_time: end datetime object
    :return: the delta datetime object
    """
    if to_time is None:
        to_time = now()

    try:
        from_time_aware = make_aware(from_time)
    except ValueError:
        from_time_aware = from_time
    try:
        to_time_aware = make_aware(to_time)
    except ValueError:
        to_time_aware = to_time

    return to_time_aware - from_time_aware


def datetime_to_isoformat(value):
    """Convert a datetime object to string representations in ISO format.

    :param value: a datetime object
    """
    if value is None:
        return None
    return value.isoformat()


def isoformat_to_datetime(value):
    """Convert string representation of a datetime in ISO format to a datetime object.

    :param value: a ISO format string representation of a datetime object
    """
    if value is None:
        return None
    return parser.parse(value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Classes and methods for Django specific backend entities"""
from typing import Generic, Set, TypeVar

from aiida.common.lang import type_check
from aiida.storage.psql_dos.models.base import Base

from . import utils

ModelType = TypeVar('ModelType')  # pylint: disable=invalid-name
SelfType = TypeVar('SelfType', bound='SqlaModelEntity')


class SqlaModelEntity(Generic[ModelType]):
    """A mixin that adds some common SQLA backend entity methods"""

    MODEL_CLASS = None
    _model: utils.ModelWrapper

    @classmethod
    def _class_check(cls):
        """Assert that the class is correctly configured"""
        assert issubclass(cls.MODEL_CLASS, Base), 'Must set the MODEL_CLASS in the derived class to a SQLA model'

    @classmethod
    def from_dbmodel(cls, dbmodel, backend):
        """Create an AiiDA Entity from the corresponding SQLA ORM model and storage backend

        :param dbmodel: the SQLAlchemy model to create the entity from
        :param backend: the corresponding storage backend
        :return: the AiiDA entity
        """
        from ..backend import PsqlDosBackend  # pylint: disable=cyclic-import
        cls._class_check()
        type_check(dbmodel, cls.MODEL_CLASS)
        type_check(backend, PsqlDosBackend)
        entity = cls.__new__(cls)
        super(SqlaModelEntity, entity).__init__(backend)
        entity._model = utils.ModelWrapper(dbmodel, backend)  # pylint: disable=protected-access
        return entity

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._class_check()

    @property
    def model(self) -> utils.ModelWrapper:
        """Return an ORM model that correctly updates and flushes the data when getting or setting a field."""
        return self._model

    @property
    def bare_model(self):
        """Return the underlying SQLA ORM model for this entity.

        .. warning:: Getting/setting attributes on this model bypasses AiiDA's internal update/flush mechanisms.
        """
        return self.model._model  # pylint: disable=protected-access

    @property
    def id(self) -> int:  # pylint: disable=redefined-builtin, invalid-name
        """
        Get the id of this entity

        :return: the entity id
        """
        return self.model.id

    @property
    def is_stored(self) -> bool:
        """
        Is this entity stored?

        :return: True if stored, False otherwise
        """
        return self.model.id is not None

    def store(self: SelfType) -> SelfType:
        """
        Store this entity

        :return: the entity itself
        """
        self.model.save()
        return self

    def _flush_if_stored(self, fields: Set[str]) -> None:
        if self.model.is_saved():
            self.model._flush(fields)  # pylint: disable=protected-access



# -*- coding: utf-8 -*-
"""Abstract utility class that helps to import calculation jobs completed outside of AiiDA."""
from abc import ABC, abstractmethod
from typing import Dict, Union

from aiida.orm import Node, RemoteData

__all__ = ('CalcJobImporter',)


class CalcJobImporter(ABC):
    """An abstract class, to define an importer for computations completed outside of AiiDA.

    This class is used to import the results of a calculation that was completed outside of AiiDA.
    The importer is responsible for parsing the output files of the calculation and creating the
    corresponding AiiDA nodes.
    """

    @staticmethod
    @abstractmethod
    def parse_remote_data(remote_data: RemoteData, **kwargs) -> Dict[str, Union[Node, Dict]]:
        """Parse the input nodes from the files in the provided ``RemoteData``.

        :param remote_data: the remote data node containing the raw input files.
        :param kwargs: additional keyword arguments to control the parsing process.
        :returns: a dictionary with the parsed inputs nodes that match the input spec of the associated ``CalcJob``.
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Change of the built in calculation entry points.

The built in calculation plugins `arithmetic.add` and `templatereplacer` have been moved and their entry point
renamed. In the change the `simpleplugins` namespace was dropped so we migrate the existing nodes.

Revision ID: django_0019
Revises: django_0018

"""
from alembic import op

revision = 'django_0019'
down_revision = 'django_0018'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.execute(
        """
        UPDATE db_dbnode SET type = 'calculation.job.arithmetic.add.ArithmeticAddCalculation.'
        WHERE type = 'calculation.job.simpleplugins.arithmetic.add.ArithmeticAddCalculation.';

        UPDATE db_dbnode SET type = 'calculation.job.templatereplacer.TemplatereplacerCalculation.'
        WHERE type = 'calculation.job.simpleplugins.templatereplacer.TemplatereplacerCalculation.';

        UPDATE db_dbnode SET process_type = 'aiida.calculations:arithmetic.add'
        WHERE process_type = 'aiida.calculations:simpleplugins.arithmetic.add';

        UPDATE db_dbnode SET process_type = 'aiida.calculations:templatereplacer'
        WHERE process_type = 'aiida.calculations:simpleplugins.templatereplacer';

        UPDATE db_dbattribute AS a SET tval = 'arithmetic.add'
        FROM db_dbnode AS n WHERE a.dbnode_id = n.id
            AND a.key = 'input_plugin'
            AND a.tval = 'simpleplugins.arithmetic.add'
            AND n.type = 'data.code.Code.';

        UPDATE db_dbattribute AS a SET tval = 'templatereplacer'
        FROM db_dbnode AS n WHERE a.dbnode_id = n.id
            AND a.key = 'input_plugin'
            AND a.tval = 'simpleplugins.templatereplacer'
            AND n.type = 'data.code.Code.';
        """
    )


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0019.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""This module contains the SQLAlchemy models for the SQLite backend.

These models are intended to be identical to those of the `psql_dos` backend,
except for changes to the database specific types:

- UUID -> CHAR(32)
- DateTime -> TZDateTime
- JSONB -> JSON

Also, `varchar_pattern_ops` indexes are not possible in sqlite.
"""
from datetime import datetime
import functools
from typing import Any, Optional, Set, Tuple

import pytz
import sqlalchemy as sa
from sqlalchemy import orm as sa_orm
from sqlalchemy.dialects.postgresql import JSONB, UUID
from sqlalchemy.dialects.sqlite import JSON

from aiida.orm.entities import EntityTypes
# we need to import all models, to ensure they are loaded on the SQLA Metadata
from aiida.storage.psql_dos.models import authinfo, base, comment, computer, group, log, node, user


class SqliteModel:
    """Represent a row in an sqlite database table"""

    def __repr__(self) -> str:
        """Return a representation of the row columns"""
        string = f'<{self.__class__.__name__}'
        for col in self.__table__.columns:  # type: ignore[attr-defined] # pylint: disable=no-member
            # don't include columns with potentially large values
            if isinstance(col.type, (JSON, sa.Text)):
                continue
            string += f' {col.name}={getattr(self, col.name)}'
        return string + '>'


class TZDateTime(sa.TypeDecorator):  # pylint: disable=abstract-method
    """A timezone naive UTC ``DateTime`` implementation for SQLite.

    see: https://docs.sqlalchemy.org/en/14/core/custom_types.html#store-timezone-aware-timestamps-as-timezone-naive-utc
    """
    impl = sa.DateTime
    cache_ok = True

    def process_bind_param(self, value: Optional[datetime], dialect):
        """Process before writing to database."""
        if value is None:
            return value
        if value.tzinfo is None:
            value = value.astimezone(pytz.utc)
        value = value.astimezone(pytz.utc).replace(tzinfo=None)
        return value

    def process_result_value(self, value: Optional[datetime], dialect):
        """Process when returning from database."""
        if value is None:
            return value
        if value.tzinfo is None:
            return value.replace(tzinfo=pytz.utc)
        return value.astimezone(pytz.utc)


SqliteBase = sa.orm.declarative_base(
    cls=SqliteModel, name='SqliteModel', metadata=sa.MetaData(naming_convention=dict(base.naming_convention))
)


def pg_to_sqlite(pg_table: sa.Table):
    """Convert a model intended for PostGreSQL to one compatible with SQLite"""
    new = pg_table.to_metadata(SqliteBase.metadata)
    for column in new.columns:
        if isinstance(column.type, UUID):
            column.type = sa.String(32)
        elif isinstance(column.type, sa.DateTime):
            column.type = TZDateTime()
        elif isinstance(column.type, JSONB):
            column.type = JSON()
    # remove any postgresql specific indexes, e.g. varchar_pattern_ops
    new.indexes.difference_update([idx for idx in new.indexes if idx.dialect_kwargs])
    return new


def create_orm_cls(klass: base.Base) -> SqliteBase:
    """Create an ORM class from an existing table in the declarative meta"""
    tbl = SqliteBase.metadata.tables[klass.__tablename__]
    return type(  # type: ignore[return-value]
        klass.__name__,
        (SqliteBase,),
        {
            '__doc__': klass.__doc__,
            '__tablename__': tbl.name,
            '__table__': tbl,
            **{col.name if col.name != 'metadata' else '_metadata': col for col in tbl.columns},
        },
    )


for table in base.Base.metadata.sorted_tables:
    pg_to_sqlite(table)

DbUser = create_orm_cls(user.DbUser)
DbComputer = create_orm_cls(computer.DbComputer)
DbAuthInfo = create_orm_cls(authinfo.DbAuthInfo)
DbGroup = create_orm_cls(group.DbGroup)
DbNode = create_orm_cls(node.DbNode)
DbGroupNodes = create_orm_cls(group.DbGroupNode)
DbComment = create_orm_cls(comment.DbComment)
DbLog = create_orm_cls(log.DbLog)
DbLink = create_orm_cls(node.DbLink)

# to-do ideally these relationships should be auto-generated in `create_orm_cls`, but this proved difficult
DbAuthInfo.aiidauser = sa_orm.relationship(  # type: ignore[attr-defined]
    'DbUser', backref=sa_orm.backref('authinfos', passive_deletes=True, cascade='all, delete')
)
DbAuthInfo.dbcomputer = sa_orm.relationship(  # type: ignore[attr-defined]
    'DbComputer', backref=sa_orm.backref('authinfos', passive_deletes=True, cascade='all, delete')
)
DbComment.dbnode = sa_orm.relationship('DbNode', backref='dbcomments')  # type: ignore[attr-defined]
DbComment.user = sa_orm.relationship('DbUser')  # type: ignore[attr-defined]
DbGroup.user = sa_orm.relationship(  # type: ignore[attr-defined]
    'DbUser', backref=sa_orm.backref('dbgroups', cascade='merge')
)
DbGroup.dbnodes = sa_orm.relationship(  # type: ignore[attr-defined]
    'DbNode', secondary='db_dbgroup_dbnodes', backref='dbgroups', lazy='dynamic'
)
DbLog.dbnode = sa_orm.relationship(  # type: ignore[attr-defined]
    'DbNode', backref=sa_orm.backref('dblogs', passive_deletes='all', cascade='merge')
)
DbNode.dbcomputer = sa_orm.relationship(  # type: ignore[attr-defined]
    'DbComputer', backref=sa_orm.backref('dbnodes', passive_deletes='all', cascade='merge')
)
DbNode.user = sa_orm.relationship('DbUser', backref=sa_orm.backref(  # type: ignore[attr-defined]
    'dbnodes',
    passive_deletes='all',
    cascade='merge',
))


@functools.lru_cache(maxsize=10)
def get_model_from_entity(entity_type: EntityTypes) -> Tuple[Any, Set[str]]:
    """Return the Sqlalchemy model and column names corresponding to the given entity."""
    model = {
        EntityTypes.USER: DbUser,
        EntityTypes.AUTHINFO: DbAuthInfo,
        EntityTypes.GROUP: DbGroup,
        EntityTypes.NODE: DbNode,
        EntityTypes.COMMENT: DbComment,
        EntityTypes.COMPUTER: DbComputer,
        EntityTypes.LOG: DbLog,
        EntityTypes.LINK: DbLink,
        EntityTypes.GROUP_NODE: DbGroupNodes
    }[entity_type]
    mapper = sa.inspect(model).mapper
    column_names = {col.name for col in mapper.c.values()}
    return model, column_names



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=cyclic-import
"""Futures that can poll or receive broadcasted messages while waiting for a task to be completed."""
import asyncio
from typing import Optional, Union

import kiwipy

from aiida.orm import Node, load_node

__all__ = ('ProcessFuture',)


class ProcessFuture(asyncio.Future):
    """Future that waits for a process to complete using both polling and listening for broadcast events if possible."""

    _filtered = None

    def __init__(
        self,
        pk: int,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        poll_interval: Union[None, int, float] = None,
        communicator: Optional[kiwipy.Communicator] = None
    ):
        """Construct a future for a process node being finished.

        If a None poll_interval is supplied polling will not be used.
        If a communicator is supplied it will be used to listen for broadcast messages.

        :param pk: process pk
        :param loop: An event loop
        :param poll_interval: optional polling interval, if None, polling is not activated.
        :param communicator: optional communicator, if None, will not subscribe to broadcasts.
        """
        from .process import ProcessState

        # create future in specified event loop
        loop = loop if loop is not None else asyncio.get_event_loop()
        super().__init__(loop=loop)

        assert not (poll_interval is None and communicator is None), 'Must poll or have a communicator to use'

        node = load_node(pk=pk)

        if node.is_terminated:
            self.set_result(node)
        else:
            self._communicator = communicator
            self.add_done_callback(lambda _: self.cleanup())

            # Try setting up a filtered broadcast subscriber
            if self._communicator is not None:

                def _subscriber(*args, **kwargs):  # pylint: disable=unused-argument
                    if not self.done():
                        self.set_result(node)

                broadcast_filter = kiwipy.BroadcastFilter(_subscriber, sender=pk)
                for state in [ProcessState.FINISHED, ProcessState.KILLED, ProcessState.EXCEPTED]:
                    broadcast_filter.add_subject_filter(f'state_changed.*.{state.value}')
                self._broadcast_identifier = self._communicator.add_broadcast_subscriber(broadcast_filter)

            # Start polling
            if poll_interval is not None:
                loop.create_task(self._poll_process(node, poll_interval))

    def cleanup(self) -> None:
        """Clean up the future by removing broadcast subscribers from the communicator if it still exists."""
        if self._communicator is not None:
            self._communicator.remove_broadcast_subscriber(self._broadcast_identifier)
            self._communicator = None
            self._broadcast_identifier = None

    async def _poll_process(self, node: Node, poll_interval: Union[int, float]) -> None:
        """Poll whether the process node has reached a terminal state."""
        while not self.done() and not node.is_terminated:
            await asyncio.sleep(poll_interval)

        if not self.done():
            self.set_result(node)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Initial main branch schema

This revision is compatible with the heads of the django and sqlalchemy branches.

Revision ID: main_0001
Revises:
Create Date: 2021-02-02

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

revision = 'main_0001'
down_revision = None
branch_labels = ('main',)
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.create_table(
        'db_dbcomputer',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False, unique=True),
        sa.Column('label', sa.String(length=255), nullable=False, unique=True),
        sa.Column('hostname', sa.String(length=255), nullable=False),
        sa.Column('description', sa.Text(), nullable=False),
        sa.Column('scheduler_type', sa.String(length=255), nullable=False),
        sa.Column('transport_type', sa.String(length=255), nullable=False),
        sa.Column('metadata', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
    )
    op.create_index(
        'ix_pat_db_dbcomputer_label',
        'db_dbcomputer', ['label'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'label': 'varchar_pattern_ops'}
    )
    op.create_table(
        'db_dbsetting',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('key', sa.String(length=1024), nullable=False, unique=True),
        sa.Column('val', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column('description', sa.Text(), nullable=False),
        sa.Column('time', sa.DateTime(timezone=True), nullable=False),
    )
    op.create_index(
        'ix_pat_db_dbsetting_key',
        'db_dbsetting',
        ['key'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'key': 'varchar_pattern_ops'},
    )
    op.create_table(
        'db_dbuser',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('email', sa.String(length=254), nullable=False, unique=True),
        sa.Column('first_name', sa.String(length=254), nullable=False),
        sa.Column('last_name', sa.String(length=254), nullable=False),
        sa.Column('institution', sa.String(length=254), nullable=False),
    )
    op.create_index(
        'ix_pat_db_dbuser_email',
        'db_dbuser',
        ['email'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'email': 'varchar_pattern_ops'},
    )
    op.create_table(
        'db_dbauthinfo',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('aiidauser_id', sa.Integer(), nullable=False, index=True),
        sa.Column('dbcomputer_id', sa.Integer(), nullable=False, index=True),
        sa.Column('metadata', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column('auth_params', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column('enabled', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ['aiidauser_id'],
            ['db_dbuser.id'],
            ondelete='CASCADE',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['dbcomputer_id'],
            ['db_dbcomputer.id'],
            ondelete='CASCADE',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.UniqueConstraint('aiidauser_id', 'dbcomputer_id'),
    )
    op.create_table(
        'db_dbgroup',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False, unique=True),
        sa.Column('label', sa.String(length=255), nullable=False, index=True),
        sa.Column('type_string', sa.String(length=255), nullable=False, index=True),
        sa.Column('time', sa.DateTime(timezone=True), nullable=False),
        sa.Column('description', sa.Text(), nullable=False),
        sa.Column('extras', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False, index=True),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['db_dbuser.id'],
            ondelete='CASCADE',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.UniqueConstraint('label', 'type_string'),
    )
    op.create_index(
        'ix_pat_db_dbgroup_label',
        'db_dbgroup',
        ['label'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'label': 'varchar_pattern_ops'},
    )
    op.create_index(
        'ix_pat_db_dbgroup_type_string',
        'db_dbgroup',
        ['type_string'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'type_string': 'varchar_pattern_ops'},
    )

    op.create_table(
        'db_dbnode',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False, unique=True),
        sa.Column('node_type', sa.String(length=255), nullable=False, index=True),
        sa.Column('process_type', sa.String(length=255), nullable=True, index=True),
        sa.Column('label', sa.String(length=255), nullable=False, index=True),
        sa.Column('description', sa.Text(), nullable=False),
        sa.Column('ctime', sa.DateTime(timezone=True), nullable=False, index=True),
        sa.Column('mtime', sa.DateTime(timezone=True), nullable=False, index=True),
        sa.Column('attributes', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column('extras', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column('repository_metadata', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column('dbcomputer_id', sa.Integer(), nullable=True, index=True),
        sa.Column('user_id', sa.Integer(), nullable=False, index=True),
        sa.ForeignKeyConstraint(
            ['dbcomputer_id'],
            ['db_dbcomputer.id'],
            ondelete='RESTRICT',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['db_dbuser.id'],
            ondelete='restrict',
            initially='DEFERRED',
            deferrable=True,
        ),
    )
    op.create_index(
        'ix_pat_db_dbnode_label',
        'db_dbnode',
        ['label'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'label': 'varchar_pattern_ops'},
    )
    op.create_index(
        'ix_pat_db_dbnode_process_type',
        'db_dbnode',
        ['process_type'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'process_type': 'varchar_pattern_ops'},
    )
    op.create_index(
        'ix_pat_db_dbnode_node_type',
        'db_dbnode',
        ['node_type'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'node_type': 'varchar_pattern_ops'},
    )

    op.create_table(
        'db_dbcomment',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False, unique=True),
        sa.Column('dbnode_id', sa.Integer(), nullable=False, index=True),
        sa.Column('ctime', sa.DateTime(timezone=True), nullable=False),
        sa.Column('mtime', sa.DateTime(timezone=True), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False, index=True),
        sa.Column('content', sa.Text(), nullable=False),
        sa.ForeignKeyConstraint(
            ['dbnode_id'],
            ['db_dbnode.id'],
            ondelete='CASCADE',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['db_dbuser.id'],
            ondelete='CASCADE',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbgroup_dbnodes',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('dbnode_id', sa.Integer(), nullable=False, index=True),
        sa.Column('dbgroup_id', sa.Integer(), nullable=False, index=True),
        sa.ForeignKeyConstraint(['dbgroup_id'], ['db_dbgroup.id'], initially='DEFERRED', deferrable=True),
        sa.ForeignKeyConstraint(['dbnode_id'], ['db_dbnode.id'], initially='DEFERRED', deferrable=True),
        sa.UniqueConstraint('dbgroup_id', 'dbnode_id'),
    )
    op.create_table(
        'db_dblink',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('input_id', sa.Integer(), nullable=False, index=True),
        sa.Column('output_id', sa.Integer(), nullable=False, index=True),
        sa.Column('label', sa.String(length=255), nullable=False, index=True),
        sa.Column('type', sa.String(length=255), nullable=False, index=True),
        sa.ForeignKeyConstraint(['input_id'], ['db_dbnode.id'], initially='DEFERRED', deferrable=True),
        sa.ForeignKeyConstraint(
            ['output_id'],
            ['db_dbnode.id'],
            ondelete='CASCADE',
            initially='DEFERRED',
            deferrable=True,
        ),
    )
    op.create_index(
        'ix_pat_db_dblink_label',
        'db_dblink',
        ['label'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'label': 'varchar_pattern_ops'},
    )
    op.create_index(
        'ix_pat_db_dblink_type',
        'db_dblink',
        ['type'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'type': 'varchar_pattern_ops'},
    )

    op.create_table(
        'db_dblog',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False, unique=True),
        sa.Column('time', sa.DateTime(timezone=True), nullable=False),
        sa.Column('loggername', sa.String(length=255), nullable=False, index=True),
        sa.Column('levelname', sa.String(length=50), nullable=False, index=True),
        sa.Column('dbnode_id', sa.Integer(), nullable=False, index=True),
        sa.Column('message', sa.Text(), nullable=False),
        sa.Column('metadata', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.ForeignKeyConstraint(
            ['dbnode_id'],
            ['db_dbnode.id'],
            ondelete='CASCADE',
            initially='DEFERRED',
            deferrable=True,
        ),
    )
    op.create_index(
        'ix_pat_db_dblog_levelname',
        'db_dblog',
        ['levelname'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'levelname': 'varchar_pattern_ops'},
    )
    op.create_index(
        'ix_pat_db_dblog_loggername',
        'db_dblog',
        ['loggername'],
        unique=False,
        postgresql_using='btree',
        postgresql_ops={'loggername': 'varchar_pattern_ops'},
    )


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of main_0001.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for all common top level AiiDA entity classes and methods"""
import abc
import copy
from enum import Enum
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Optional, Protocol, Type, TypeVar, cast

from plumpy.base.utils import call_with_super_check, super_check

from aiida.common import exceptions
from aiida.common.lang import classproperty, type_check
from aiida.manage import get_manager

if TYPE_CHECKING:
    from aiida.orm.implementation import BackendEntity, StorageBackend
    from aiida.orm.querybuilder import FilterType, OrderByType, QueryBuilder

__all__ = ('Entity', 'Collection', 'EntityAttributesMixin', 'EntityExtrasMixin', 'EntityTypes')

CollectionType = TypeVar('CollectionType', bound='Collection')
EntityType = TypeVar('EntityType', bound='Entity')
BackendEntityType = TypeVar('BackendEntityType', bound='BackendEntity')

_NO_DEFAULT: Any = tuple()


class EntityTypes(Enum):
    """Enum for referring to ORM entities in a backend-agnostic manner."""
    AUTHINFO = 'authinfo'
    COMMENT = 'comment'
    COMPUTER = 'computer'
    GROUP = 'group'
    LOG = 'log'
    NODE = 'node'
    USER = 'user'
    LINK = 'link'
    GROUP_NODE = 'group_node'


class Collection(abc.ABC, Generic[EntityType]):
    """Container class that represents the collection of objects of a particular entity type."""

    @staticmethod
    @abc.abstractmethod
    def _entity_base_cls() -> Type[EntityType]:
        """The allowed entity class or subclasses thereof."""

    @classmethod
    @lru_cache(maxsize=100)
    def get_cached(cls, entity_class: Type[EntityType], backend: 'StorageBackend'):
        """Get the cached collection instance for the given entity class and backend.

        :param backend: the backend instance to get the collection for
        """
        from aiida.orm.implementation import StorageBackend
        type_check(backend, StorageBackend)
        return cls(entity_class, backend=backend)

    def __init__(self, entity_class: Type[EntityType], backend: Optional['StorageBackend'] = None) -> None:
        """ Construct a new entity collection.

        :param entity_class: the entity type e.g. User, Computer, etc
        :param backend: the backend instance to get the collection for, or use the default
        """
        from aiida.orm.implementation import StorageBackend
        type_check(backend, StorageBackend, allow_none=True)
        assert issubclass(entity_class, self._entity_base_cls())
        self._backend = backend or get_manager().get_profile_storage()
        self._entity_type = entity_class

    def __call__(self: CollectionType, backend: 'StorageBackend') -> CollectionType:
        """Get or create a cached collection using a new backend."""
        if backend is self._backend:
            return self
        return self.get_cached(self.entity_type, backend=backend)  # type: ignore

    @property
    def entity_type(self) -> Type[EntityType]:
        """The entity type for this instance."""
        return self._entity_type

    @property
    def backend(self) -> 'StorageBackend':
        """Return the backend."""
        return self._backend

    def query(
        self,
        filters: Optional['FilterType'] = None,
        order_by: Optional['OrderByType'] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None
    ) -> 'QueryBuilder':
        """Get a query builder for the objects of this collection.

        :param filters: the keyword value pair filters to match
        :param order_by: a list of (key, direction) pairs specifying the sort order
        :param limit: the maximum number of results to return
        :param offset: number of initial results to be skipped
        """
        from . import querybuilder

        filters = filters or {}
        order_by = {self.entity_type: order_by} if order_by else {}

        query = querybuilder.QueryBuilder(backend=self._backend, limit=limit, offset=offset)
        query.append(self.entity_type, project='*', filters=filters)
        query.order_by([order_by])
        return query

    def get(self, **filters: Any) -> EntityType:
        """Get a single collection entry that matches the filter criteria.

        :param filters: the filters identifying the object to get

        :return: the entry
        """
        res = self.query(filters=filters)
        return res.one()[0]

    def find(
        self,
        filters: Optional['FilterType'] = None,
        order_by: Optional['OrderByType'] = None,
        limit: Optional[int] = None
    ) -> List[EntityType]:
        """Find collection entries matching the filter criteria.

        :param filters: the keyword value pair filters to match
        :param order_by: a list of (key, direction) pairs specifying the sort order
        :param limit: the maximum number of results to return

        :return: a list of resulting matches
        """
        query = self.query(filters=filters, order_by=order_by, limit=limit)
        return cast(List[EntityType], query.all(flat=True))

    def all(self) -> List[EntityType]:
        """Get all entities in this collection.

        :return: A list of all entities
        """
        return cast(List[EntityType], self.query().all(flat=True))  # pylint: disable=no-member

    def count(self, filters: Optional['FilterType'] = None) -> int:
        """Count entities in this collection according to criteria.

        :param filters: the keyword value pair filters to match

        :return: The number of entities found using the supplied criteria
        """
        return self.query(filters=filters).count()


class Entity(abc.ABC, Generic[BackendEntityType]):
    """An AiiDA entity"""

    @classproperty
    @abc.abstractmethod
    def objects(cls: EntityType) -> Collection[EntityType]:  # pylint: disable=no-self-argument,disable=no-self-use
        """Get a collection for objects of this type, with the default backend.

        :return: an object that can be used to access entities of this type
        """

    @classmethod
    def get(cls, **kwargs):
        return cls.objects.get(**kwargs)  # pylint: disable=no-member

    @classmethod
    def from_backend_entity(cls: Type[EntityType], backend_entity: BackendEntityType) -> EntityType:
        """
        Construct an entity from a backend entity instance

        :param backend_entity: the backend entity

        :return: an AiiDA entity instance
        """
        from .implementation.entities import BackendEntity

        type_check(backend_entity, BackendEntity)
        entity = cls.__new__(cls)
        entity._backend_entity = backend_entity
        call_with_super_check(entity.initialize)
        return entity

    def __init__(self, backend_entity: BackendEntityType) -> None:
        """
        :param backend_entity: the backend model supporting this entity
        """
        self._backend_entity = backend_entity
        call_with_super_check(self.initialize)

    @super_check
    def initialize(self) -> None:
        """Initialize instance attributes.

        This will be called after the constructor is called or an entity is created from an existing backend entity.
        """

    @property
    def id(self) -> int:  # pylint: disable=invalid-name
        """Return the id for this entity.

        This identifier is guaranteed to be unique amongst entities of the same type for a single backend instance.

        :return: the entity's id
        """
        return self._backend_entity.id

    @property
    def pk(self) -> int:
        """Return the primary key for this entity.

        This identifier is guaranteed to be unique amongst entities of the same type for a single backend instance.

        :return: the entity's principal key
        """
        return self.id

    def store(self: EntityType) -> EntityType:
        """Store the entity."""
        self._backend_entity.store()
        return self

    @property
    def is_stored(self) -> bool:
        """Return whether the entity is stored."""
        return self._backend_entity.is_stored

    @property
    def backend(self) -> 'StorageBackend':
        """Get the backend for this entity"""
        return self._backend_entity.backend

    @property
    def backend_entity(self) -> BackendEntityType:
        """Get the implementing class for this object"""
        return self._backend_entity


class EntityProtocol(Protocol):
    """Protocol for attributes required by Entity mixins."""

    @property
    def backend_entity(self) -> 'BackendEntity':
        ...

    @property
    def is_stored(self) -> bool:
        ...


class EntityAttributesMixin:
    """Mixin class that adds all methods for the attributes column to an entity."""

    @property
    def attributes(self: EntityProtocol) -> Dict[str, Any]:
        """Return the complete attributes dictionary.

        .. warning:: While the entity is unstored, this will return references of the attributes on the database model,
            meaning that changes on the returned values (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            attributes will be a deep copy and mutations of the database attributes will have to go through the
            appropriate set methods. Therefore, once stored, retrieving a deep copy can be a heavy operation. If you
            only need the keys or some values, use the iterators `attributes_keys` and `attributes_items`, or the
            getters `get_attribute` and `get_attribute_many` instead.

        :return: the attributes as a dictionary
        """
        attributes = self.backend_entity.attributes

        if self.is_stored:
            attributes = copy.deepcopy(attributes)

        return attributes

    def get_attribute(self: EntityProtocol, key: str, default=_NO_DEFAULT) -> Any:
        """Return the value of an attribute.

        .. warning:: While the entity is unstored, this will return a reference of the attribute on the database model,
            meaning that changes on the returned value (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            attribute will be a deep copy and mutations of the database attributes will have to go through the
            appropriate set methods.

        :param key: name of the attribute
        :param default: return this value instead of raising if the attribute does not exist
        :return: the value of the attribute
        :raises AttributeError: if the attribute does not exist and no default is specified
        """
        try:
            attribute = self.backend_entity.get_attribute(key)
        except AttributeError:
            if default is _NO_DEFAULT:
                raise
            attribute = default

        if self.is_stored:
            attribute = copy.deepcopy(attribute)

        return attribute

    def get_attribute_many(self: EntityProtocol, keys: List[str]) -> List[Any]:
        """Return the values of multiple attributes.

        .. warning:: While the entity is unstored, this will return references of the attributes on the database model,
            meaning that changes on the returned values (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            attributes will be a deep copy and mutations of the database attributes will have to go through the
            appropriate set methods. Therefore, once stored, retrieving a deep copy can be a heavy operation. If you
            only need the keys or some values, use the iterators `attributes_keys` and `attributes_items`, or the
            getters `get_attribute` and `get_attribute_many` instead.

        :param keys: a list of attribute names
        :return: a list of attribute values
        :raises AttributeError: if at least one attribute does not exist
        """
        attributes = self.backend_entity.get_attribute_many(keys)

        if self.is_stored:
            attributes = copy.deepcopy(attributes)

        return attributes

    def set_attribute(self: EntityProtocol, key: str, value: Any) -> None:
        """Set an attribute to the given value.

        :param key: name of the attribute
        :param value: value of the attribute
        :raise aiida.common.ValidationError: if the key is invalid, i.e. contains periods
        :raise aiida.common.ModificationNotAllowed: if the entity is stored
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.set_attribute(key, value)

    def set_attribute_many(self: EntityProtocol, attributes: Dict[str, Any]) -> None:
        """Set multiple attributes.

        .. note:: This will override any existing attributes that are present in the new dictionary.

        :param attributes: a dictionary with the attributes to set
        :raise aiida.common.ValidationError: if any of the keys are invalid, i.e. contain periods
        :raise aiida.common.ModificationNotAllowed: if the entity is stored
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.set_attribute_many(attributes)

    def reset_attributes(self: EntityProtocol, attributes: Dict[str, Any]) -> None:
        """Reset the attributes.

        .. note:: This will completely clear any existing attributes and replace them with the new dictionary.

        :param attributes: a dictionary with the attributes to set
        :raise aiida.common.ValidationError: if any of the keys are invalid, i.e. contain periods
        :raise aiida.common.ModificationNotAllowed: if the entity is stored
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.reset_attributes(attributes)

    def delete_attribute(self: EntityProtocol, key: str) -> None:
        """Delete an attribute.

        :param key: name of the attribute
        :raises AttributeError: if the attribute does not exist
        :raise aiida.common.ModificationNotAllowed: if the entity is stored
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.delete_attribute(key)

    def delete_attribute_many(self: EntityProtocol, keys: List[str]) -> None:
        """Delete multiple attributes.

        :param keys: names of the attributes to delete
        :raises AttributeError: if at least one of the attribute does not exist
        :raise aiida.common.ModificationNotAllowed: if the entity is stored
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.delete_attribute_many(keys)

    def clear_attributes(self: EntityProtocol) -> None:
        """Delete all attributes."""
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

        self.backend_entity.clear_attributes()

    def attributes_items(self: EntityProtocol):
        """Return an iterator over the attributes.

        :return: an iterator with attribute key value pairs
        """
        return self.backend_entity.attributes_items()

    def attributes_keys(self: EntityProtocol):
        """Return an iterator over the attribute keys.

        :return: an iterator with attribute keys
        """
        return self.backend_entity.attributes_keys()


class EntityExtrasMixin:
    """Mixin class that adds all methods for the extras column to an entity."""

    @property
    def extras(self: EntityProtocol) -> Dict[str, Any]:
        """Return the complete extras dictionary.

        .. warning:: While the entity is unstored, this will return references of the extras on the database model,
            meaning that changes on the returned values (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            extras will be a deep copy and mutations of the database extras will have to go through the appropriate set
            methods. Therefore, once stored, retrieving a deep copy can be a heavy operation. If you only need the keys
            or some values, use the iterators `extras_keys` and `extras_items`, or the getters `get_extra` and
            `get_extra_many` instead.

        :return: the extras as a dictionary
        """
        extras = self.backend_entity.extras

        if self.is_stored:
            extras = copy.deepcopy(extras)

        return extras

    def get_extra(self: EntityProtocol, key: str, default: Any = _NO_DEFAULT) -> Any:
        """Return the value of an extra.

        .. warning:: While the entity is unstored, this will return a reference of the extra on the database model,
            meaning that changes on the returned value (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            extra will be a deep copy and mutations of the database extras will have to go through the appropriate set
            methods.

        :param key: name of the extra
        :param default: return this value instead of raising if the attribute does not exist
        :return: the value of the extra
        :raises AttributeError: if the extra does not exist and no default is specified
        """
        try:
            extra = self.backend_entity.get_extra(key)
        except AttributeError:
            if default is _NO_DEFAULT:
                raise
            extra = default

        if self.is_stored:
            extra = copy.deepcopy(extra)

        return extra

    def get_extra_many(self: EntityProtocol, keys: List[str]) -> List[Any]:
        """Return the values of multiple extras.

        .. warning:: While the entity is unstored, this will return references of the extras on the database model,
            meaning that changes on the returned values (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            extras will be a deep copy and mutations of the database extras will have to go through the appropriate set
            methods. Therefore, once stored, retrieving a deep copy can be a heavy operation. If you only need the keys
            or some values, use the iterators `extras_keys` and `extras_items`, or the getters `get_extra` and
            `get_extra_many` instead.

        :param keys: a list of extra names
        :return: a list of extra values
        :raises AttributeError: if at least one extra does not exist
        """
        extras = self.backend_entity.get_extra_many(keys)

        if self.is_stored:
            extras = copy.deepcopy(extras)

        return extras

    def set_extra(self: EntityProtocol, key: str, value: Any) -> None:
        """Set an extra to the given value.

        :param key: name of the extra
        :param value: value of the extra
        :raise aiida.common.ValidationError: if the key is invalid, i.e. contains periods
        """
        self.backend_entity.set_extra(key, value)

    def set_extra_many(self: EntityProtocol, extras: Dict[str, Any]) -> None:
        """Set multiple extras.

        .. note:: This will override any existing extras that are present in the new dictionary.

        :param extras: a dictionary with the extras to set
        :raise aiida.common.ValidationError: if any of the keys are invalid, i.e. contain periods
        """
        self.backend_entity.set_extra_many(extras)

    def reset_extras(self: EntityProtocol, extras: Dict[str, Any]) -> None:
        """Reset the extras.

        .. note:: This will completely clear any existing extras and replace them with the new dictionary.

        :param extras: a dictionary with the extras to set
        :raise aiida.common.ValidationError: if any of the keys are invalid, i.e. contain periods
        """
        self.backend_entity.reset_extras(extras)

    def delete_extra(self: EntityProtocol, key: str) -> None:
        """Delete an extra.

        :param key: name of the extra
        :raises AttributeError: if the extra does not exist
        """
        self.backend_entity.delete_extra(key)

    def delete_extra_many(self: EntityProtocol, keys: List[str]) -> None:
        """Delete multiple extras.

        :param keys: names of the extras to delete
        :raises AttributeError: if at least one of the extra does not exist
        """
        self.backend_entity.delete_extra_many(keys)

    def clear_extras(self: EntityProtocol) -> None:
        """Delete all extras."""
        self.backend_entity.clear_extras()

    def extras_items(self: EntityProtocol):
        """Return an iterator over the extras.

        :return: an iterator with extra key value pairs
        """
        return self.backend_entity.extras_items()

    def extras_keys(self: EntityProtocol):
        """Return an iterator over the extra keys.

        :return: an iterator with extra keys
        """
        return self.backend_entity.extras_keys()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Initial django schema

Revision ID: django_0001
Revises:
Create Date: 2017-06-28 17:12:23.327195

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

revision = 'django_0001'
down_revision = None
branch_labels = ('django',)
depends_on = None


def upgrade():
    """Migrations for the upgrade."""

    # dummy django tables
    op.create_table(
        'auth_group',
        sa.Column('id', sa.INTEGER(), nullable=False, primary_key=True),
    )
    op.create_table(
        'auth_group_permissions',
        sa.Column('id', sa.INTEGER(), nullable=False, primary_key=True),
    )
    op.create_table(
        'auth_permission',
        sa.Column('id', sa.INTEGER(), nullable=False, primary_key=True),
    )
    op.create_table(
        'django_content_type',
        sa.Column('id', sa.INTEGER(), nullable=False, primary_key=True),
    )
    op.create_table(
        'django_migrations',
        sa.Column('id', sa.INTEGER(), nullable=False, primary_key=True),
    )

    op.create_table(
        'db_dbuser',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbuser_pkey'),
        sa.Column('email', sa.VARCHAR(length=75), nullable=False),
        sa.Column('password', sa.VARCHAR(length=128), nullable=False),
        sa.Column('is_superuser', sa.BOOLEAN(), nullable=False),
        sa.Column('first_name', sa.VARCHAR(length=254), nullable=False),
        sa.Column('last_name', sa.VARCHAR(length=254), nullable=False),
        sa.Column('institution', sa.VARCHAR(length=254), nullable=False),
        sa.Column('is_staff', sa.BOOLEAN(), nullable=False),
        sa.Column('is_active', sa.BOOLEAN(), nullable=False),
        sa.Column('last_login', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('date_joined', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.UniqueConstraint('email', name='db_dbuser_email_key'),
        sa.Index(
            'db_dbuser_email_30150b7e_like',
            'email',
            postgresql_using='btree',
            postgresql_ops={'email': 'varchar_pattern_ops'},
        ),
    )

    op.create_table(
        'db_dbcomputer',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbcomputer_pkey'),
        sa.Column('uuid', sa.VARCHAR(length=36), nullable=False),
        sa.Column('name', sa.VARCHAR(length=255), nullable=False),
        sa.Column('hostname', sa.VARCHAR(length=255), nullable=False),
        sa.Column('description', sa.TEXT(), nullable=False),
        sa.Column('enabled', sa.BOOLEAN(), nullable=False),
        sa.Column('transport_type', sa.VARCHAR(length=255), nullable=False),
        sa.Column('scheduler_type', sa.VARCHAR(length=255), nullable=False),
        sa.Column('transport_params', sa.TEXT(), nullable=False),
        sa.Column('metadata', sa.TEXT(), nullable=False),
        sa.UniqueConstraint('name', name='db_dbcomputer_name_key'),
        sa.Index(
            'db_dbcomputer_name_f1800b1a_like',
            'name',
            postgresql_using='btree',
            postgresql_ops={'name': 'varchar_pattern_ops'},
        ),
    )

    op.create_table(
        'db_dbgroup',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbgroup_pkey'),
        sa.Column('uuid', sa.VARCHAR(length=36), nullable=False),
        sa.Column('name', sa.VARCHAR(length=255), nullable=False),
        sa.Column('type', sa.VARCHAR(length=255), nullable=False),
        sa.Column('time', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('description', sa.TEXT(), nullable=False),
        sa.Column('user_id', sa.INTEGER(), nullable=False),
        sa.UniqueConstraint('name', 'type', name='db_dbgroup_name_type_12656f33_uniq'),
        sa.Index('db_dbgroup_name_66c75272', 'name'),
        sa.Index('db_dbgroup_type_23b2a748', 'type'),
        sa.Index('db_dbgroup_user_id_100f8a51', 'user_id'),
        sa.Index(
            'db_dbgroup_name_66c75272_like',
            'name',
            postgresql_using='btree',
            postgresql_ops={'name': 'varchar_pattern_ops'},
        ),
        sa.Index(
            'db_dbgroup_type_23b2a748_like',
            'type',
            postgresql_using='btree',
            postgresql_ops={'type': 'varchar_pattern_ops'},
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['db_dbuser.id'],
            name='db_dbgroup_user_id_100f8a51_fk_db_dbuser_id',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dblock',
        sa.Column('key', sa.VARCHAR(length=255), nullable=False),
        sa.PrimaryKeyConstraint('key', name='db_dblock_pkey'),
        sa.Column('creation', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('timeout', sa.INTEGER(), nullable=False),
        sa.Column('owner', sa.VARCHAR(length=255), nullable=False),
        sa.Index(
            'db_dblock_key_048c6767_like',
            'key',
            postgresql_using='btree',
            postgresql_ops={'key': 'varchar_pattern_ops'},
        ),
    )

    op.create_table(
        'db_dblog',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dblog_pkey'),
        sa.Column('time', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('loggername', sa.VARCHAR(length=255), nullable=False),
        sa.Column('levelname', sa.VARCHAR(length=50), nullable=False),
        sa.Column('objname', sa.VARCHAR(length=255), nullable=False),
        sa.Column('objpk', sa.INTEGER(), nullable=True),
        sa.Column('message', sa.TEXT(), nullable=False),
        sa.Column('metadata', sa.TEXT(), nullable=False),
        sa.Index('db_dblog_levelname_ad5dc346', 'levelname'),
        sa.Index('db_dblog_loggername_00b5ba16', 'loggername'),
        sa.Index('db_dblog_objname_69932b1e', 'objname'),
        sa.Index('db_dblog_objpk_fc47afa9', 'objpk'),
        sa.Index(
            'db_dblog_levelname_ad5dc346_like',
            'levelname',
            postgresql_using='btree',
            postgresql_ops={'levelname': 'varchar_pattern_ops'},
        ),
        sa.Index(
            'db_dblog_loggername_00b5ba16_like',
            'loggername',
            postgresql_using='btree',
            postgresql_ops={'loggername': 'varchar_pattern_ops'},
        ),
        sa.Index(
            'db_dblog_objname_69932b1e_like',
            'objname',
            postgresql_using='btree',
            postgresql_ops={'objname': 'varchar_pattern_ops'},
        ),
    )

    op.create_table(
        'db_dbnode',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbnode_pkey'),
        sa.Column('uuid', sa.VARCHAR(length=36), nullable=False),
        sa.Column('type', sa.VARCHAR(length=255), nullable=False),
        sa.Column('label', sa.VARCHAR(length=255), nullable=False),
        sa.Column('description', sa.TEXT(), nullable=False),
        sa.Column('ctime', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('mtime', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('nodeversion', sa.INTEGER(), nullable=False),
        sa.Column('public', sa.BOOLEAN(), nullable=False),
        sa.Column('dbcomputer_id', sa.INTEGER(), nullable=True),
        sa.Column('user_id', sa.INTEGER(), nullable=False),
        sa.Index('db_dbnode_dbcomputer_id_315372a3', 'dbcomputer_id'),
        sa.Index('db_dbnode_label_6469539e', 'label'),
        sa.Index('db_dbnode_type_a8ce9753', 'type'),
        sa.Index('db_dbnode_user_id_12e7aeaf', 'user_id'),
        sa.Index(
            'db_dbnode_label_6469539e_like',
            'label',
            postgresql_using='btree',
            postgresql_ops={'label': 'varchar_pattern_ops'},
        ),
        sa.Index(
            'db_dbnode_type_a8ce9753_like',
            'type',
            postgresql_using='btree',
            postgresql_ops={'type': 'varchar_pattern_ops'},
        ),
        sa.ForeignKeyConstraint(
            ['dbcomputer_id'],
            ['db_dbcomputer.id'],
            name='db_dbnode_dbcomputer_id_315372a3_fk_db_dbcomputer_id',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['db_dbuser.id'],
            name='db_dbnode_user_id_12e7aeaf_fk_db_dbuser_id',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbattribute',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbattribute_pkey'),
        sa.Column('datatype', sa.VARCHAR(length=10), nullable=False),
        sa.Column('dbnode_id', sa.INTEGER(), nullable=False),
        sa.Column('key', sa.VARCHAR(length=1024), nullable=False),
        sa.Column('bval', sa.BOOLEAN(), nullable=True),
        sa.Column('ival', sa.INTEGER(), nullable=True),
        sa.Column('fval', sa.FLOAT(), nullable=True),
        sa.Column('tval', sa.TEXT(), nullable=False),
        sa.Column('dval', postgresql.TIMESTAMP(timezone=True), nullable=True),
        sa.UniqueConstraint('dbnode_id', 'key', name='db_dbattribute_dbnode_id_key_c589e447_uniq'),
        sa.Index('db_dbattribute_datatype_91c4dc04', 'datatype'),
        sa.Index('db_dbattribute_dbnode_id_253bf153', 'dbnode_id'),
        sa.Index('db_dbattribute_key_ac2bc4e4', 'key'),
        sa.Index(
            'db_dbattribute_datatype_91c4dc04_like',
            'datatype',
            postgresql_using='btree',
            postgresql_ops={'datatype': 'varchar_pattern_ops'},
        ),
        sa.Index(
            'db_dbattribute_key_ac2bc4e4_like',
            'key',
            postgresql_using='btree',
            postgresql_ops={'key': 'varchar_pattern_ops'},
        ),
        sa.ForeignKeyConstraint(
            ['dbnode_id'],
            ['db_dbnode.id'],
            name='db_dbattribute_dbnode_id_253bf153_fk_db_dbnode_id',
            deferrable=True,
            initially='DEFERRED',
        ),
    )

    op.create_table(
        'db_dbextra',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbextra_pkey'),
        sa.Column('datatype', sa.VARCHAR(length=10), nullable=False),
        sa.Column('dbnode_id', sa.INTEGER(), nullable=False),
        sa.Column('key', sa.VARCHAR(length=1024), nullable=False),
        sa.Column('bval', sa.BOOLEAN(), nullable=True),
        sa.Column('ival', sa.INTEGER(), nullable=True),
        sa.Column('fval', sa.FLOAT(), nullable=True),
        sa.Column('tval', sa.TEXT(), nullable=False),
        sa.Column('dval', postgresql.TIMESTAMP(timezone=True), nullable=True),
        sa.UniqueConstraint('dbnode_id', 'key', name='db_dbextra_dbnode_id_key_aa56fd37_uniq'),
        sa.Index('db_dbextra_datatype_2eba38c6', 'datatype'),
        sa.Index('db_dbextra_dbnode_id_c7fe8961', 'dbnode_id'),
        sa.Index('db_dbextra_key_b1a8abc6', 'key'),
        sa.Index(
            'db_dbextra_datatype_2eba38c6_like',
            'datatype',
            postgresql_using='btree',
            postgresql_ops={'datatype': 'varchar_pattern_ops'},
        ),
        sa.Index(
            'db_dbextra_key_b1a8abc6_like',
            'key',
            postgresql_using='btree',
            postgresql_ops={'key': 'varchar_pattern_ops'},
        ),
        sa.ForeignKeyConstraint(
            ['dbnode_id'],
            ['db_dbnode.id'],
            name='db_dbextra_dbnode_id_c7fe8961_fk_db_dbnode_id',
            deferrable=True,
            initially='DEFERRED',
        ),
    )

    op.create_table(
        'db_dblink',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dblink_pkey'),
        sa.Column('input_id', sa.INTEGER(), nullable=False),
        sa.Column('output_id', sa.INTEGER(), nullable=False),
        sa.Column('label', sa.VARCHAR(length=255), nullable=False),
        sa.UniqueConstraint('input_id', 'output_id', name='db_dblink_input_id_output_id_fbe99cb5_uniq'),
        sa.UniqueConstraint('output_id', 'label', name='db_dblink_output_id_label_00bdb9c7_uniq'),
        sa.Index('db_dblink_input_id_9245bd73', 'input_id'),
        sa.Index('db_dblink_label_f1343cfb', 'label'),
        sa.Index('db_dblink_output_id_c0167528', 'output_id'),
        sa.Index(
            'db_dblink_label_f1343cfb_like',
            'label',
            postgresql_using='btree',
            postgresql_ops={'label': 'varchar_pattern_ops'},
        ),
        sa.ForeignKeyConstraint(
            ['input_id'],
            ['db_dbnode.id'],
            name='db_dblink_input_id_9245bd73_fk_db_dbnode_id',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['output_id'],
            ['db_dbnode.id'],
            name='db_dblink_output_id_c0167528_fk_db_dbnode_id',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbgroup_dbnodes',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbgroup_dbnodes_pkey'),
        sa.Column('dbnode_id', sa.INTEGER(), nullable=False),
        sa.Column('dbgroup_id', sa.INTEGER(), nullable=False),
        sa.UniqueConstraint('dbgroup_id', 'dbnode_id', name='db_dbgroup_dbnodes_dbgroup_id_dbnode_id_eee23cce_uniq'),
        sa.Index('db_dbgroup_dbnodes_dbgroup_id_9d3a0f9d', 'dbgroup_id'),
        sa.Index('db_dbgroup_dbnodes_dbnode_id_118b9439', 'dbnode_id'),
        sa.ForeignKeyConstraint(
            ['dbgroup_id'],
            ['db_dbgroup.id'],
            name='db_dbgroup_dbnodes_dbgroup_id_9d3a0f9d_fk_db_dbgroup_id',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['dbnode_id'],
            ['db_dbnode.id'],
            name='db_dbgroup_dbnodes_dbnode_id_118b9439_fk_db_dbnode_id',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbcalcstate',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbcalcstate_pkey'),
        sa.Column('dbnode_id', sa.INTEGER(), nullable=False),
        sa.Column('state', sa.VARCHAR(length=25), nullable=False),
        sa.Column('time', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.UniqueConstraint('dbnode_id', 'state', name='db_dbcalcstate_dbnode_id_state_b4a14db3_uniq'),
        sa.Index('db_dbcalcstate_dbnode_id_f217a84c', 'dbnode_id'),
        sa.Index('db_dbcalcstate_state_0bf54584', 'state'),
        sa.Index(
            'db_dbcalcstate_state_0bf54584_like',
            'state',
            postgresql_using='btree',
            postgresql_ops={'state': 'varchar_pattern_ops'},
        ),
        sa.ForeignKeyConstraint(
            ['dbnode_id'],
            ['db_dbnode.id'],
            name='db_dbcalcstate_dbnode_id_f217a84c_fk_db_dbnode_id',
            deferrable=True,
            initially='DEFERRED',
        ),
    )

    op.create_table(
        'db_dbcomment',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbcomment_pkey'),
        sa.Column('uuid', sa.VARCHAR(length=36), nullable=False),
        sa.Column('dbnode_id', sa.INTEGER(), nullable=False),
        sa.Column('ctime', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('mtime', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('user_id', sa.INTEGER(), nullable=False),
        sa.Column('content', sa.TEXT(), nullable=False),
        sa.Index('db_dbcomment_dbnode_id_3b812b6b', 'dbnode_id'),
        sa.Index('db_dbcomment_user_id_8ed5e360', 'user_id'),
        sa.ForeignKeyConstraint(
            ['dbnode_id'],
            ['db_dbnode.id'],
            name='db_dbcomment_dbnode_id_3b812b6b_fk_db_dbnode_id',
            deferrable=True,
            initially='DEFERRED',
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['db_dbuser.id'],
            name='db_dbcomment_user_id_8ed5e360_fk_db_dbuser_id',
            deferrable=True,
            initially='DEFERRED',
        ),
    )

    op.create_table(
        'db_dbpath',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbpath_pkey'),
        sa.Column('parent_id', sa.INTEGER(), nullable=False),
        sa.Column('child_id', sa.INTEGER(), nullable=False),
        sa.Column('depth', sa.INTEGER(), nullable=False),
        sa.Column('entry_edge_id', sa.INTEGER(), nullable=True),
        sa.Column('direct_edge_id', sa.INTEGER(), nullable=True),
        sa.Column('exit_edge_id', sa.INTEGER(), nullable=True),
        sa.Index('db_dbpath_child_id_d8228636', 'child_id'),
        sa.Index('db_dbpath_parent_id_3b82d6c8', 'parent_id'),
        sa.ForeignKeyConstraint(
            ['child_id'],
            ['db_dbnode.id'],
            name='db_dbpath_child_id_d8228636_fk_db_dbnode_id',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['parent_id'],
            ['db_dbnode.id'],
            name='db_dbpath_parent_id_3b82d6c8_fk_db_dbnode_id',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbsetting',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbsetting_pkey'),
        sa.Column('key', sa.VARCHAR(length=1024), nullable=False),
        sa.Column('description', sa.TEXT(), nullable=False),
        sa.Column('time', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('datatype', sa.VARCHAR(length=10), nullable=False),
        sa.Column('bval', sa.BOOLEAN(), nullable=True),
        sa.Column('ival', sa.INTEGER(), nullable=True),
        sa.Column('fval', sa.FLOAT(), nullable=True),
        sa.Column('tval', sa.TEXT(), nullable=False),
        sa.Column('dval', postgresql.TIMESTAMP(timezone=True), nullable=True),
        sa.UniqueConstraint('key', name='db_dbsetting_key_1b84beb4_uniq'),
        sa.Index('db_dbsetting_datatype_49f4397c', 'datatype'),
        sa.Index('db_dbsetting_key_1b84beb4', 'key'),
        sa.Index(
            'db_dbsetting_datatype_49f4397c_like',
            'datatype',
            postgresql_using='btree',
            postgresql_ops={'datatype': 'varchar_pattern_ops'},
        ),
        sa.Index(
            'db_dbsetting_key_1b84beb4_like',
            'key',
            postgresql_using='btree',
            postgresql_ops={'key': 'varchar_pattern_ops'},
        ),
    )

    op.create_table(
        'db_dbuser_groups',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbuser_groups_pkey'),
        sa.Column('dbuser_id', sa.INTEGER(), nullable=False),
        sa.Column('group_id', sa.INTEGER(), nullable=False),
        sa.UniqueConstraint('dbuser_id', 'group_id', name='db_dbuser_groups_dbuser_id_group_id_9155eb4f_uniq'),
        sa.Index('db_dbuser_groups_dbuser_id_480b3520', 'dbuser_id'),
        sa.Index('db_dbuser_groups_group_id_8478d87e', 'group_id'),
        sa.ForeignKeyConstraint(
            ['dbuser_id'],
            ['db_dbuser.id'],
            name='db_dbuser_groups_dbuser_id_480b3520_fk_db_dbuser_id',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['group_id'],
            ['auth_group.id'],
            name='db_dbuser_groups_group_id_8478d87e_fk_auth_group_id',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbuser_user_permissions',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbuser_user_permissions_pkey'),
        sa.Column('dbuser_id', sa.INTEGER(), nullable=False),
        sa.Column('permission_id', sa.INTEGER(), nullable=False),
        sa.UniqueConstraint(
            'dbuser_id', 'permission_id', name='db_dbuser_user_permissio_dbuser_id_permission_id_e6cbabe4_uniq'
        ),
        sa.Index('db_dbuser_user_permissions_dbuser_id_364456ee', 'dbuser_id'),
        sa.Index('db_dbuser_user_permissions_permission_id_c5aafc54', 'permission_id'),
        sa.ForeignKeyConstraint(
            ['dbuser_id'],
            ['db_dbuser.id'],
            name='db_dbuser_user_permissions_dbuser_id_364456ee_fk_db_dbuser_id',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['permission_id'],
            ['auth_permission.id'],
            name='db_dbuser_user_permi_permission_id_c5aafc54_fk_auth_perm',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbworkflow',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbworkflow_pkey'),
        sa.Column('uuid', sa.VARCHAR(length=36), nullable=False),
        sa.Column('ctime', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('mtime', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('user_id', sa.INTEGER(), nullable=False),
        sa.Column('label', sa.VARCHAR(length=255), nullable=False),
        sa.Column('description', sa.TEXT(), nullable=False),
        sa.Column('nodeversion', sa.INTEGER(), nullable=False),
        sa.Column('lastsyncedversion', sa.INTEGER(), nullable=False),
        sa.Column('state', sa.VARCHAR(length=255), nullable=False),
        sa.Column('report', sa.TEXT(), nullable=False),
        sa.Column('module', sa.TEXT(), nullable=False),
        sa.Column('module_class', sa.TEXT(), nullable=False),
        sa.Column('script_path', sa.TEXT(), nullable=False),
        sa.Column('script_md5', sa.VARCHAR(length=255), nullable=False),
        sa.Index('db_dbworkflow_label_7368f34a', 'label'),
        sa.Index('db_dbworkflow_user_id_ef1f3251', 'user_id'),
        sa.Index(
            'db_dbworkflow_label_7368f34a_like',
            'label',
            postgresql_using='btree',
            postgresql_ops={'label': 'varchar_pattern_ops'},
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['db_dbuser.id'],
            name='db_dbworkflow_user_id_ef1f3251_fk_db_dbuser_id',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbworkflowstep',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbworkflowstep_pkey'),
        sa.Column('parent_id', sa.INTEGER(), nullable=False),
        sa.Column('user_id', sa.INTEGER(), nullable=False),
        sa.Column('name', sa.VARCHAR(length=255), nullable=False),
        sa.Column('time', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('nextcall', sa.VARCHAR(length=255), nullable=False),
        sa.Column('state', sa.VARCHAR(length=255), nullable=False),
        sa.UniqueConstraint('parent_id', 'name', name='db_dbworkflowstep_parent_id_name_111027e3_uniq'),
        sa.Index('db_dbworkflowstep_parent_id_ffb754d9', 'parent_id'),
        sa.Index('db_dbworkflowstep_user_id_04282431', 'user_id'),
        sa.ForeignKeyConstraint(
            ['parent_id'],
            ['db_dbworkflow.id'],
            name='db_dbworkflowstep_parent_id_ffb754d9_fk_db_dbworkflow_id',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['db_dbuser.id'],
            name='db_dbworkflowstep_user_id_04282431_fk_db_dbuser_id',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbworkflowdata',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbworkflowdata_pkey'),
        sa.Column('parent_id', sa.INTEGER(), nullable=False),
        sa.Column('name', sa.VARCHAR(length=255), nullable=False),
        sa.Column('time', postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.Column('data_type', sa.VARCHAR(length=255), nullable=False),
        sa.Column('value_type', sa.VARCHAR(length=255), nullable=False),
        sa.Column('json_value', sa.TEXT(), nullable=False),
        sa.Column('aiida_obj_id', sa.INTEGER(), nullable=True),
        sa.UniqueConstraint(
            'parent_id', 'name', 'data_type', name='db_dbworkflowdata_parent_id_name_data_type_a4b50dae_uniq'
        ),
        sa.Index('db_dbworkflowdata_aiida_obj_id_70a2d33b', 'aiida_obj_id'),
        sa.Index('db_dbworkflowdata_parent_id_ff4dbf8d', 'parent_id'),
        sa.ForeignKeyConstraint(
            ['aiida_obj_id'],
            ['db_dbnode.id'],
            name='db_dbworkflowdata_aiida_obj_id_70a2d33b_fk_db_dbnode_id',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['parent_id'],
            ['db_dbworkflow.id'],
            name='db_dbworkflowdata_parent_id_ff4dbf8d_fk_db_dbworkflow_id',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbworkflowstep_calculations',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbworkflowstep_calculations_pkey'),
        sa.Column('dbworkflowstep_id', sa.INTEGER(), nullable=False),
        sa.Column('dbnode_id', sa.INTEGER(), nullable=False),
        sa.UniqueConstraint(
            'dbworkflowstep_id', 'dbnode_id', name='db_dbworkflowstep_calcul_dbworkflowstep_id_dbnode_60f50d02_uniq'
        ),
        sa.Index('db_dbworkflowstep_calculations_dbnode_id_0d07b7a7', 'dbnode_id'),
        sa.Index('db_dbworkflowstep_calculations_dbworkflowstep_id_575c3637', 'dbworkflowstep_id'),
        sa.ForeignKeyConstraint(
            ['dbnode_id'],
            ['db_dbnode.id'],
            name='db_dbworkflowstep_ca_dbnode_id_0d07b7a7_fk_db_dbnode',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['dbworkflowstep_id'],
            ['db_dbworkflowstep.id'],
            name='db_dbworkflowstep_ca_dbworkflowstep_id_575c3637_fk_db_dbwork',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbworkflowstep_sub_workflows',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbworkflowstep_sub_workflows_pkey'),
        sa.Column('dbworkflowstep_id', sa.INTEGER(), nullable=False),
        sa.Column('dbworkflow_id', sa.INTEGER(), nullable=False),
        sa.UniqueConstraint(
            'dbworkflowstep_id',
            'dbworkflow_id',
            name='db_dbworkflowstep_sub_wo_dbworkflowstep_id_dbwork_e9b2b624_uniq',
        ),
        sa.Index('db_dbworkflowstep_sub_workflows_dbworkflow_id_dca4d103', 'dbworkflow_id'),
        sa.Index('db_dbworkflowstep_sub_workflows_dbworkflowstep_id_e183bbb7', 'dbworkflowstep_id'),
        sa.ForeignKeyConstraint(
            ['dbworkflow_id'],
            ['db_dbworkflow.id'],
            name='db_dbworkflowstep_su_dbworkflow_id_dca4d103_fk_db_dbwork',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['dbworkflowstep_id'],
            ['db_dbworkflowstep.id'],
            name='db_dbworkflowstep_su_dbworkflowstep_id_e183bbb7_fk_db_dbwork',
            initially='DEFERRED',
            deferrable=True,
        ),
    )

    op.create_table(
        'db_dbauthinfo',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.PrimaryKeyConstraint('id', name='db_dbauthinfo_pkey'),
        sa.Column('aiidauser_id', sa.INTEGER(), nullable=False),
        sa.Column('dbcomputer_id', sa.INTEGER(), nullable=False),
        sa.Column('metadata', sa.TEXT(), nullable=False),
        sa.Column('auth_params', sa.TEXT(), nullable=False),
        sa.Column('enabled', sa.BOOLEAN(), nullable=False),
        sa.UniqueConstraint(
            'aiidauser_id', 'dbcomputer_id', name='db_dbauthinfo_aiidauser_id_dbcomputer_id_777cdaa8_uniq'
        ),
        sa.Index('db_dbauthinfo_aiidauser_id_0684fdfb', 'aiidauser_id'),
        sa.Index('db_dbauthinfo_dbcomputer_id_424f7ac4', 'dbcomputer_id'),
        sa.ForeignKeyConstraint(
            ['aiidauser_id'],
            ['db_dbuser.id'],
            name='db_dbauthinfo_aiidauser_id_0684fdfb_fk_db_dbuser_id',
            initially='DEFERRED',
            deferrable=True,
        ),
        sa.ForeignKeyConstraint(
            ['dbcomputer_id'],
            ['db_dbcomputer.id'],
            name='db_dbauthinfo_dbcomputer_id_424f7ac4_fk_db_dbcomputer_id',
            initially='DEFERRED',
            deferrable=True,
        ),
    )


def downgrade():
    """Migrations for the downgrade."""
    op.drop_table('db_dbauthinfo')
    op.drop_table('db_dbworkflowstep_calculations')
    op.drop_table('db_dbworkflowstep_sub_workflows')
    op.drop_table('db_dbworkflowdata')
    op.drop_table('db_dbworkflowstep')
    op.drop_table('db_dbworkflow')
    op.drop_table('db_dbuser_user_permissions')
    op.drop_table('db_dbuser_groups')
    op.drop_table('db_dbgroup_dbnodes')
    op.drop_table('db_dbgroup')
    op.drop_table('db_dblink')
    op.drop_table('db_dbpath')
    op.drop_table('db_dbcalcstate')
    op.drop_table('db_dbcomment')
    op.drop_table('db_dbattribute')
    op.drop_table('db_dbextra')
    op.drop_table('db_dbnode')
    op.drop_table('db_dbcomputer')
    op.drop_table('db_dblog')
    op.drop_table('db_dbsetting')
    op.drop_table('db_dblock')
    op.drop_table('db_dbuser')

    op.drop_table('auth_group_permissions')
    op.drop_table('auth_permission')
    op.drop_table('auth_group')
    op.drop_table('django_content_type')
    op.drop_table('django_migrations')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Reusable command line interface options for Computer commands."""
import click

from aiida.cmdline.params import options, types
from aiida.cmdline.params.options.interactive import InteractiveOption, TemplateInteractiveOption
from aiida.cmdline.params.options.overridable import OverridableOption


def should_call_default_mpiprocs_per_machine(ctx):  # pylint: disable=invalid-name
    """
    Return True if the scheduler can accept 'default_mpiprocs_per_machine',
    False otherwise.

    If there is a problem in determining the scheduler, return True to
    avoid exceptions.
    """
    from aiida.common.exceptions import ValidationError

    scheduler_ep = ctx.params['scheduler']
    if scheduler_ep is not None:
        try:
            scheduler_cls = scheduler_ep.load()
        except ImportError:
            raise ImportError(f"Unable to load the '{scheduler_ep.name}' scheduler")
    else:
        raise ValidationError(
            'The should_call_... function should always be run (and prompted) AFTER asking for a scheduler'
        )

    job_resource_cls = scheduler_cls.job_resource_class
    if job_resource_cls is None:
        # Odd situation...
        return False

    return job_resource_cls.accepts_default_mpiprocs_per_machine()


LABEL = options.LABEL.clone(
    prompt='Computer label',
    cls=InteractiveOption,
    required=True,
    help='Unique, human-readable label for this computer.'
)

HOSTNAME = options.HOSTNAME.clone(
    prompt='Hostname',
    cls=InteractiveOption,
    required=True,
    help='The fully qualified hostname of the computer (e.g. daint.cscs.ch). '
    'Use "localhost" when setting up the computer that AiiDA is running on.',
)

DESCRIPTION = options.DESCRIPTION.clone(
    prompt='Description', cls=InteractiveOption, help='A human-readable description of this computer.'
)

TRANSPORT = options.TRANSPORT.clone(prompt='Transport plugin', cls=InteractiveOption)

SCHEDULER = options.SCHEDULER.clone(prompt='Scheduler plugin', cls=InteractiveOption)

SHEBANG = OverridableOption(
    '--shebang',
    prompt='Shebang line (first line of each script, starting with #!)',
    default='#!/bin/bash',
    cls=InteractiveOption,
    help='Specify the first line of the submission script for this computer (only the bash shell is supported).',
    type=types.ShebangParamType()
)

WORKDIR = OverridableOption(
    '-w',
    '--work-dir',
    prompt='Work directory on the computer',
    default='/scratch/{username}/aiida/',
    cls=InteractiveOption,
    help='The absolute path of the directory on the computer where AiiDA will '
    'run the calculations (often a "scratch" directory).'
    'The {username} string will be replaced by your username on the remote computer.'
)

MPI_RUN_COMMAND = OverridableOption(
    '-m',
    '--mpirun-command',
    prompt='Mpirun command',
    default='mpirun -np {tot_num_mpiprocs}',
    cls=InteractiveOption,
    help='The mpirun command needed on the cluster to run parallel MPI programs. The {tot_num_mpiprocs} string will be '
    'replaced by the total number of cpus. See the scheduler docs for further scheduler-dependent template variables.',
    type=types.MpirunCommandParamType()
)

MPI_PROCS_PER_MACHINE = OverridableOption(
    '--mpiprocs-per-machine',
    prompt='Default number of CPUs per machine',
    cls=InteractiveOption,
    prompt_fn=should_call_default_mpiprocs_per_machine,
    required_fn=False,
    type=click.INT,
    help='The default number of MPI processes that should be executed per machine (node), if not otherwise specified.'
    'Use 0 to specify no default value.',
)

DEFAULT_MEMORY_PER_MACHINE = OverridableOption(
    '--default-memory-per-machine',
    prompt='Default amount of memory per machine (kB).',
    cls=InteractiveOption,
    type=click.INT,
    help='The default amount of RAM (kB) that should be allocated per machine (node), if not otherwise specified.'
)

PREPEND_TEXT = OverridableOption(
    '--prepend-text',
    cls=TemplateInteractiveOption,
    prompt='Prepend script',
    type=click.STRING,
    default='',
    help='Bash commands that should be prepended to the executable call in all submit scripts for this computer.',
    extension='.bash',
    header='PREPEND_TEXT: if there is any bash commands that should be prepended to the executable call in all '
    'submit scripts for this computer, type that between the equal signs below and save the file.',
    footer='All lines that start with `#=` will be ignored.'
)

APPEND_TEXT = OverridableOption(
    '--append-text',
    cls=TemplateInteractiveOption,
    prompt='Append script',
    type=click.STRING,
    default='',
    help='Bash commands that should be appended to the executable call in all submit scripts for this computer.',
    extension='.bash',
    header='APPEND_TEXT: if there is any bash commands that should be appended to the executable call in all '
    'submit scripts for this computer, type that between the equal signs below and save the file.',
    footer='All lines that start with `#=` will be ignored.'
)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities for dealing with links between nodes."""
from collections import OrderedDict
from collections.abc import Mapping
from typing import TYPE_CHECKING, Generator, Iterator, List, NamedTuple, Optional

from aiida.common import exceptions
from aiida.common.lang import type_check

if TYPE_CHECKING:
    from aiida.common.links import LinkType
    from aiida.orm import Node
    from aiida.orm.implementation.storage_backend import StorageBackend

__all__ = ('LinkPair', 'LinkTriple', 'LinkManager', 'validate_link')


class LinkPair(NamedTuple):
    link_type: 'LinkType'
    link_label: str


class LinkTriple(NamedTuple):
    node: 'Node'
    link_type: 'LinkType'
    link_label: str


class LinkQuadruple(NamedTuple):
    source_id: int
    target_id: int
    link_type: 'LinkType'
    link_label: str


def link_triple_exists(
    source: 'Node',
    target: 'Node',
    link_type: 'LinkType',
    link_label: str,
    backend: Optional['StorageBackend'] = None
) -> bool:
    """Return whether a link with the given type and label exists between the given source and target node.

    :param source: node from which the link is outgoing
    :param target: node to which the link is incoming
    :param link_type: the link type
    :param link_label: the link label
    :return: boolean, True if the link triple exists, False otherwise
    """
    from aiida.orm import Node, QueryBuilder

    # First check if the triple exist in the cache, in the case of an unstored target node
    if target._incoming_cache and LinkTriple(source, link_type, link_label) in target._incoming_cache:  # pylint: disable=protected-access
        return True

    # If either node is unstored (i.e. does not have a pk), the link cannot exist in the database, so no need to check
    if source.pk is None or target.pk is None:
        return False

    # Here we have two stored nodes, so we need to check if the same link already exists in the database.
    # Finding just a single match is sufficient so we can use the `limit` clause for efficiency
    builder = QueryBuilder(backend=backend)
    builder.append(Node, filters={'id': source.id}, project=['id'])
    builder.append(Node, filters={'id': target.id}, edge_filters={'type': link_type.value, 'label': link_label})
    builder.limit(1)

    return builder.count() != 0


def validate_link(
    source: 'Node',
    target: 'Node',
    link_type: 'LinkType',
    link_label: str,
    backend: Optional['StorageBackend'] = None
) -> None:
    """
    Validate adding a link of the given type and label from a given node to ourself.

    This function will first validate the class types of the inputs and will subsequently validate whether a link of
    the specified type is allowed at all between the nodes types of the source and target.

    Subsequently, the validity of the "indegree" and "outdegree" of the proposed link is validated, which means
    validating that the uniqueness constraints of the incoming links into the target node and the outgoing links from
    the source node are not violated. In AiiDA's provenance graph each link type has one of the following three types
    of "degree" character::

        * unique
        * unique pair
        * unique triple

    Each degree character has a different unique constraint on its links, here defined for the indegree::

        * unique: any target node, it can only have a single incoming link of this type, regardless of the link label.
        * unique pair: a node can have an infinite amount of incoming links of this type, as long as the labels within
            that sub set, are unique. In short, it is the link pair, i.e. the tuple of the link type and label, that has
            a uniquess constraint for the incoming links to a given node.
        * unique triple: a node can have an infinite amount of incoming links of this type, as long as the triple tuple
            of source node, link type and link label is unique. In other words, it is the link triple that has a
            uniqueness constraint for the incoming links.

    The same holds for outdegree, but then it concerns outgoing links from the source node to the target node.

    For illustration purposes, consider the following example provenance graphs that are considered legal, where
    `WN`, `DN` and `CN` represent a `WorkflowNode`, a `DataNode` and a `CalculationNode`, respectively::

                    1                    2                    3
            ______     ______          ______          ______     ______
           |      |   |      |        |      |        |      |   |      |
           |  WN  |   |  DN  |        |  DN  |        |  WN  |   |  WN  |
           |______|   |______|        |______|        |______|   |______|
                |     /                |   |               |     /
              a |    / a             a |   | b           a |    / a
               _|___/                  |___|_             _|___/
              |      |                |      |           |      |
              |  CN  |                |  CN  |           |  DN  |
              |______|                |______|           |______|

    In example 1, the link uniqueness constraint is not violated because despite the labels having the same label `a`,
    their link types, `CALL_CALC` and `INPUT_CALC`, respectively, are different and their `unique_pair` indegree is
    not violated.

    Similarly, in the second example, the constraint is not violated, because despite both links having the same link
    type `INPUT_CALC`, the have different labels, so the `unique_pair` indegree of the `INPUT_CALC` is not violated.

    Finally, in the third example, we see two `WorkflowNodes` both returning the same `DataNode` and with the same
    label. Despite the two incoming links here having both the same type as well as the same label, the uniqueness
    constraint is not violated, because the indegree for `RETURN` links is `unique_triple` which means that the triple
    of source node and link type and label should be unique.

    :param source: the node from which the link is coming
    :param target: the node to which the link is going
    :param link_type: the type of link
    :param link_label: link label
    :raise TypeError: if `source` or `target` is not a Node instance, or `link_type` is not a `LinkType` enum
    :raise ValueError: if the proposed link is invalid
    """
    # yapf: disable
    from aiida.common.links import LinkType, validate_link_label
    from aiida.orm import CalculationNode, Data, Node, WorkflowNode

    type_check(link_type, LinkType, f'link_type should be a LinkType enum but got: {type(link_type)}')
    type_check(source, Node, f'source should be a `Node` but got: {type(source)}')
    type_check(target, Node, f'target should be a `Node` but got: {type(target)}')

    if source.uuid is None or target.uuid is None:
        raise ValueError('source or target node does not have a UUID')

    if source.uuid == target.uuid:
        raise ValueError('cannot add a link to oneself')

    try:
        validate_link_label(link_label)
    except ValueError as exception:
        raise ValueError(f'invalid link label `{link_label}`: {exception}')

    # For each link type, define a tuple that defines the valid types for the source and target node, as well as
    # the outdegree and indegree character. If the degree is `unique` that means that there can only be a single
    # link of this type regardless of the label. If instead it is `unique_label`, an infinite amount of links of that
    # type can be defined, as long as the link label is unique for the sub set of links of that type. Finally, for
    # `unique_triple` the triple of node, link type and link label has to be unique.
    link_mapping = {
        LinkType.CALL_CALC: (WorkflowNode, CalculationNode, 'unique_triple', 'unique'),
        LinkType.CALL_WORK: (WorkflowNode, WorkflowNode, 'unique_triple', 'unique'),
        LinkType.CREATE: (CalculationNode, Data, 'unique_pair', 'unique'),
        LinkType.INPUT_CALC: (Data, CalculationNode, 'unique_triple', 'unique_pair'),
        LinkType.INPUT_WORK: (Data, WorkflowNode, 'unique_triple', 'unique_pair'),
        LinkType.RETURN: (WorkflowNode, Data, 'unique_pair', 'unique_triple'),
    }

    type_source, type_target, outdegree, indegree = link_mapping[link_type]

    if not isinstance(source, type_source) or not isinstance(target, type_target):
        raise ValueError(f'cannot add a {link_type} link from {type(source)} to {type(target)}')

    if outdegree == 'unique_triple' or indegree == 'unique_triple':
        # For a `unique_triple` degree we just have to check if an identical triple already exist, either in the cache
        # or stored, in which case, the new proposed link is a duplicate and thus illegal
        duplicate_link_triple = link_triple_exists(source, target, link_type, link_label, backend)

    # If the outdegree is `unique` there cannot already be any other outgoing link of that type
    if outdegree == 'unique' and source.get_outgoing(link_type=link_type, only_uuid=True).all():
        raise ValueError(f'node<{source.uuid}> already has an outgoing {link_type} link')

    # If the outdegree is `unique_pair`, then the link labels for outgoing links of this type should be unique
    elif outdegree == 'unique_pair' and source.get_outgoing(
            link_type=link_type, only_uuid=True, link_label_filter=link_label).all():
        raise ValueError(f'node<{source.uuid}> already has an outgoing {link_type} link with label "{link_label}"')

    # If the outdegree is `unique_triple`, then the link triples of link type, link label and target should be unique
    elif outdegree == 'unique_triple' and duplicate_link_triple:
        raise ValueError('node<{}> already has an outgoing {} link with label "{}" from node<{}>'.format(
            source.uuid, link_type, link_label, target.uuid))

    # If the indegree is `unique` there cannot already be any other incoming links of that type
    if indegree == 'unique' and target.get_incoming(link_type=link_type, only_uuid=True).all():
        raise ValueError(f'node<{target.uuid}> already has an incoming {link_type} link')

    # If the indegree is `unique_pair`, then the link labels for incoming links of this type should be unique
    elif indegree == 'unique_pair' and target.get_incoming(
            link_type=link_type, link_label_filter=link_label, only_uuid=True).all():
        raise ValueError(f'node<{target.uuid}> already has an incoming {link_type} link with label "{link_label}"')

    # If the indegree is `unique_triple`, then the link triples of link type, link label and source should be unique
    elif indegree == 'unique_triple' and duplicate_link_triple:
        raise ValueError('node<{}> already has an incoming {} link with label "{}" from node<{}>'.format(
            target.uuid, link_type, link_label, source.uuid))


class LinkManager:
    """
    Class to convert a list of LinkTriple tuples into an iterator.

    It defines convenience methods to retrieve certain subsets of LinkTriple while checking for consistency.
    For example::

        LinkManager.one(): returns the only entry in the list or it raises an exception
        LinkManager.first(): returns the first entry from the list
        LinkManager.all(): returns all entries from list

    The methods `all_nodes` and `all_link_labels` are syntactic sugar wrappers around `all` to get a list of only the
    incoming nodes or link labels, respectively.
    """

    def __init__(self, link_triples: List[LinkTriple]):
        """Initialise the collection."""
        self.link_triples = link_triples

    def __iter__(self) -> Iterator[LinkTriple]:
        """Return an iterator of LinkTriple instances.

        :return: iterator of LinkTriple instances
        """
        return iter(self.link_triples)

    def __next__(self) -> Generator[LinkTriple, None, None]:
        """Return the next element in the iterator.

        :return: LinkTriple
        """
        for link_triple in self.link_triples:
            yield link_triple

    def __bool__(self):
        return bool(len(self.link_triples))

    def next(self) -> Generator[LinkTriple, None, None]:
        """Return the next element in the iterator.

        :return: LinkTriple
        """
        return self.__next__()

    def one(self) -> LinkTriple:
        """Return a single entry from the iterator.

        If the iterator contains no or more than one entry, an exception will be raised
        :return: LinkTriple instance
        :raises ValueError: if the iterator contains anything but one entry
        """
        if self.link_triples:
            if len(self.link_triples) > 1:
                raise ValueError('more than one entry found')
            return self.link_triples[0]

        raise ValueError('no entries found')

    def first(self) -> Optional[LinkTriple]:
        """Return the first entry from the iterator.

        :return: LinkTriple instance or None if no entries were matched
        """
        if self.link_triples:
            return self.link_triples[0]

        return None

    def all(self) -> List[LinkTriple]:
        """Return all entries from the list.

        :return: list of LinkTriple instances
        """
        return self.link_triples

    def all_nodes(self) -> List['Node']:
        """Return a list of all nodes.

        :return: list of nodes
        """
        return [entry.node for entry in self.all()]

    def all_link_pairs(self) -> List[LinkPair]:
        """Return a list of all link pairs.

        :return: list of LinkPair instances
        """
        return [LinkPair(entry.link_type, entry.link_label) for entry in self.all()]

    def all_link_labels(self) -> List[str]:
        """Return a list of all link labels.

        :return: list of link labels
        """
        return [entry.link_label for entry in self.all()]

    def get_node_by_label(self, label: str) -> 'Node':
        """Return the node from list for given label.

        :return: node that corresponds to the given label
        :raises aiida.common.NotExistent: if the label is not present among the link_triples
        """
        matching_entry = None
        for entry in self.link_triples:
            if entry.link_label == label:
                if matching_entry is None:
                    matching_entry = entry.node
                else:
                    raise exceptions.MultipleObjectsError(
                        f'more than one neighbor with the label {label} found'
                    )

        if matching_entry is None:
            raise exceptions.NotExistent(f'no neighbor with the label {label} found')

        return matching_entry

    def nested(self, sort=True):
        """Construct (nested) dictionary of matched nodes that mirrors the original nesting of link namespaces.

        Process input and output namespaces can be nested, however the link labels that represent them in the database
        have a flat hierarchy, and so the link labels are flattened representations of the nested namespaces.
        This function reconstructs the original node nesting based on the flattened links.

        :return: dictionary of nested namespaces
        :raises KeyError: if there are duplicate link labels in a namespace
        """
        from aiida.engine.processes.ports import PORT_NAMESPACE_SEPARATOR

        nested: dict = {}

        for entry in self.link_triples:

            current_namespace = nested
            breadcrumbs = entry.link_label.split(PORT_NAMESPACE_SEPARATOR)

            # The last element is the "leaf" port name the preceding elements are nested port namespaces
            port_name = breadcrumbs[-1]
            port_namespaces = breadcrumbs[:-1]

            # Get the nested namespace
            for subspace in port_namespaces:
                current_namespace = current_namespace.setdefault(subspace, {})

            # Insert the node at the given port name
            if port_name in current_namespace:
                raise KeyError(f"duplicate label '{port_name}' in namespace '{'.'.join(port_namespaces)}'")

            current_namespace[port_name] = entry.node

        if sort:
            return OrderedDict(sorted(nested.items(), key=lambda x: (not isinstance(x[1], Mapping), x)))

        return nested



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with `Node` sub class for processes."""

import enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union

from plumpy.process_states import ProcessState

from aiida.common.lang import classproperty
from aiida.common.links import LinkType
from aiida.orm.utils.mixins import Sealable

from ..node import Node

if TYPE_CHECKING:
    from aiida.engine.processes import Process
    from aiida.engine.processes.builder import ProcessBuilder

__all__ = ('ProcessNode',)


class ProcessNode(Sealable, Node):
    """
    Base class for all nodes representing the execution of a process

    This class and its subclasses serve as proxies in the database, for actual `Process` instances being run. The
    `Process` instance in memory will leverage an instance of this class (the exact sub class depends on the sub class
    of `Process`) to persist important information of its state to the database. This serves as a way for the user to
    inspect the state of the `Process` during its execution as well as a permanent record of its execution in the
    provenance graph, after the execution has terminated.
    """
    # pylint: disable=too-many-public-methods,abstract-method

    CHECKPOINT_KEY = 'checkpoints'
    EXCEPTION_KEY = 'exception'
    EXIT_MESSAGE_KEY = 'exit_message'
    EXIT_STATUS_KEY = 'exit_status'
    PROCESS_PAUSED_KEY = 'paused'
    PROCESS_LABEL_KEY = 'process_label'
    PROCESS_STATE_KEY = 'process_state'
    PROCESS_STATUS_KEY = 'process_status'

    # The link_type might not be correct while the object is being created.
    _hash_ignored_inputs = ['CALL_CALC', 'CALL_WORK']

    _unstorable_message = 'only Data, WorkflowNode, CalculationNode or their subclasses can be stored'

    def __str__(self) -> str:
        base = super().__str__()
        if self.process_type:
            return f'{base} ({self.process_type})'

        return f'{base}'

    @classproperty
    def _updatable_attributes(cls) -> Tuple[str, ...]:
        # pylint: disable=no-self-argument
        return super()._updatable_attributes + (
            cls.PROCESS_PAUSED_KEY,
            cls.CHECKPOINT_KEY,
            cls.EXCEPTION_KEY,
            cls.EXIT_MESSAGE_KEY,
            cls.EXIT_STATUS_KEY,
            cls.PROCESS_LABEL_KEY,
            cls.PROCESS_STATE_KEY,
            cls.PROCESS_STATUS_KEY,
        )

    @property
    def logger(self):
        """
        Get the logger of the Calculation object, so that it also logs to the DB.

        :return: LoggerAdapter object, that works like a logger, but also has the 'extra' embedded
        """
        from aiida.orm.utils.log import create_logger_adapter
        return create_logger_adapter(self._logger, self)

    def get_builder_restart(self) -> 'ProcessBuilder':
        """Return a `ProcessBuilder` that is ready to relaunch the process that created this node.

        The process class will be set based on the `process_type` of this node and the inputs of the builder will be
        prepopulated with the inputs registered for this node. This functionality is very useful if a process has
        completed and you want to relaunch it with slightly different inputs.

        :return: `~aiida.engine.processes.builder.ProcessBuilder` instance
        """
        builder = self.process_class.get_builder()
        builder._update(self.get_incoming(link_type=(LinkType.INPUT_CALC, LinkType.INPUT_WORK)).nested())  # pylint: disable=protected-access

        return builder

    @property
    def process_class(self) -> Type['Process']:
        """Return the process class that was used to create this node.

        :return: `Process` class
        :raises ValueError: if no process type is defined, it is an invalid process type string or cannot be resolved
            to load the corresponding class
        """
        from aiida.common import exceptions
        from aiida.plugins.entry_point import load_entry_point_from_string

        if not self.process_type:
            raise ValueError(f'no process type for Node<{self.pk}>: cannot recreate process class')

        try:
            process_class = load_entry_point_from_string(self.process_type)
        except exceptions.EntryPointError as exception:
            raise ValueError(
                f'could not load process class for entry point `{self.process_type}` for Node<{self.pk}>: {exception}'
            )
        except ValueError:
            try:
                import importlib
                module_name, class_name = self.process_type.rsplit('.', 1)
                module = importlib.import_module(module_name)
                process_class = getattr(module, class_name)
            except (AttributeError, ValueError, ImportError) as exception:
                raise ValueError(
                    f'could not load process class from `{self.process_type}` for Node<{self.pk}>: {exception}'
                )

        return process_class

    def set_process_type(self, process_type_string: str) -> None:
        """
        Set the process type string.

        :param process_type: the process type string identifying the class using this process node as storage.
        """
        self.process_type = process_type_string

    @property
    def process_label(self) -> Optional[str]:
        """
        Return the process label

        :returns: the process label
        """
        return self.get_attribute(self.PROCESS_LABEL_KEY, None)

    def set_process_label(self, label: str) -> None:
        """
        Set the process label

        :param label: process label string
        """
        self.set_attribute(self.PROCESS_LABEL_KEY, label)

    @property
    def process_state(self) -> Optional[ProcessState]:
        """
        Return the process state

        :returns: the process state instance of ProcessState enum
        """
        state = self.get_attribute(self.PROCESS_STATE_KEY, None)

        if state is None:
            return state

        return ProcessState(state)

    def set_process_state(self, state: Union[str, ProcessState]):
        """
        Set the process state

        :param state: value or instance of ProcessState enum
        """
        if isinstance(state, ProcessState):
            state = state.value
        return self.set_attribute(self.PROCESS_STATE_KEY, state)

    @property
    def process_status(self) -> Optional[str]:
        """
        Return the process status

        The process status is a generic status message e.g. the reason it might be paused or when it is being killed

        :returns: the process status
        """
        return self.get_attribute(self.PROCESS_STATUS_KEY, None)

    def set_process_status(self, status: Optional[str]) -> None:
        """
        Set the process status

        The process status is a generic status message e.g. the reason it might be paused or when it is being killed.
        If status is None, the corresponding attribute will be deleted.

        :param status: string process status
        """
        if status is None:
            try:
                self.delete_attribute(self.PROCESS_STATUS_KEY)
            except AttributeError:
                pass
            return None

        if not isinstance(status, str):
            raise TypeError('process status should be a string')

        return self.set_attribute(self.PROCESS_STATUS_KEY, status)

    @property
    def is_terminated(self) -> bool:
        """
        Return whether the process has terminated

        Terminated means that the process has reached any terminal state.

        :return: True if the process has terminated, False otherwise
        :rtype: bool
        """
        return self.is_excepted or self.is_finished or self.is_killed

    @property
    def is_excepted(self) -> bool:
        """
        Return whether the process has excepted

        Excepted means that during execution of the process, an exception was raised that was not caught.

        :return: True if during execution of the process an exception occurred, False otherwise
        :rtype: bool
        """
        return self.process_state == ProcessState.EXCEPTED

    @property
    def is_killed(self) -> bool:
        """
        Return whether the process was killed

        Killed means the process was killed directly by the user or by the calling process being killed.

        :return: True if the process was killed, False otherwise
        :rtype: bool
        """
        return self.process_state == ProcessState.KILLED

    @property
    def is_finished(self) -> bool:
        """
        Return whether the process has finished

        Finished means that the process reached a terminal state nominally.
        Note that this does not necessarily mean successfully, but there were no exceptions and it was not killed.

        :return: True if the process has finished, False otherwise
        :rtype: bool
        """
        return self.process_state == ProcessState.FINISHED

    @property
    def is_finished_ok(self) -> bool:
        """
        Return whether the process has finished successfully

        Finished successfully means that it terminated nominally and had a zero exit status.

        :return: True if the process has finished successfully, False otherwise
        :rtype: bool
        """
        return self.is_finished and self.exit_status == 0

    @property
    def is_failed(self) -> bool:
        """
        Return whether the process has failed

        Failed means that the process terminated nominally but it had a non-zero exit status.

        :return: True if the process has failed, False otherwise
        :rtype: bool
        """
        return self.is_finished and self.exit_status != 0

    @property
    def exit_status(self) -> Optional[int]:
        """
        Return the exit status of the process

        :returns: the exit status, an integer exit code or None
        """
        return self.get_attribute(self.EXIT_STATUS_KEY, None)

    def set_exit_status(self, status: Union[None, enum.Enum, int]) -> None:
        """
        Set the exit status of the process

        :param state: an integer exit code or None, which will be interpreted as zero
        """
        if status is None:
            status = 0

        if isinstance(status, enum.Enum):
            status = status.value

        if not isinstance(status, int):
            raise ValueError(f'exit status has to be an integer, got {status}')

        return self.set_attribute(self.EXIT_STATUS_KEY, status)

    @property
    def exit_message(self) -> Optional[str]:
        """
        Return the exit message of the process

        :returns: the exit message
        """
        return self.get_attribute(self.EXIT_MESSAGE_KEY, None)

    def set_exit_message(self, message: Optional[str]) -> None:
        """
        Set the exit message of the process, if None nothing will be done

        :param message: a string message
        """
        if message is None:
            return None

        if not isinstance(message, str):
            raise ValueError(f'exit message has to be a string type, got {type(message)}')

        return self.set_attribute(self.EXIT_MESSAGE_KEY, message)

    @property
    def exception(self) -> Optional[str]:
        """
        Return the exception of the process or None if the process is not excepted.

        If the process is marked as excepted yet there is no exception attribute, an empty string will be returned.

        :returns: the exception message or None
        """
        if self.is_excepted:
            return self.get_attribute(self.EXCEPTION_KEY, '')

        return None

    def set_exception(self, exception: str) -> None:
        """
        Set the exception of the process

        :param exception: the exception message
        """
        if not isinstance(exception, str):
            raise ValueError(f'exception message has to be a string type, got {type(exception)}')

        return self.set_attribute(self.EXCEPTION_KEY, exception)

    @property
    def checkpoint(self) -> Optional[Dict[str, Any]]:
        """
        Return the checkpoint bundle set for the process

        :returns: checkpoint bundle if it exists, None otherwise
        """
        return self.get_attribute(self.CHECKPOINT_KEY, None)

    def set_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        """
        Set the checkpoint bundle set for the process

        :param state: string representation of the stepper state info
        """
        return self.set_attribute(self.CHECKPOINT_KEY, checkpoint)

    def delete_checkpoint(self) -> None:
        """
        Delete the checkpoint bundle set for the process
        """
        try:
            self.delete_attribute(self.CHECKPOINT_KEY)
        except AttributeError:
            pass

    @property
    def paused(self) -> bool:
        """
        Return whether the process is paused

        :returns: True if the Calculation is marked as paused, False otherwise
        """
        return self.get_attribute(self.PROCESS_PAUSED_KEY, False)

    def pause(self) -> None:
        """
        Mark the process as paused by setting the corresponding attribute.

        This serves only to reflect that the corresponding Process is paused and so this method should not be called
        by anyone but the Process instance itself.
        """
        return self.set_attribute(self.PROCESS_PAUSED_KEY, True)

    def unpause(self) -> None:
        """
        Mark the process as unpaused by removing the corresponding attribute.

        This serves only to reflect that the corresponding Process is unpaused and so this method should not be called
        by anyone but the Process instance itself.
        """
        try:
            self.delete_attribute(self.PROCESS_PAUSED_KEY)
        except AttributeError:
            pass

    @property
    def called(self) -> List['ProcessNode']:
        """
        Return a list of nodes that the process called

        :returns: list of process nodes called by this process
        """
        return self.get_outgoing(link_type=(LinkType.CALL_CALC, LinkType.CALL_WORK)).all_nodes()

    @property
    def called_descendants(self) -> List['ProcessNode']:
        """
        Return a list of all nodes that have been called downstream of this process

        This will recursively find all the called processes for this process and its children.
        """
        descendants = []

        for descendant in self.called:
            descendants.append(descendant)
            descendants.extend(descendant.called_descendants)

        return descendants

    @property
    def caller(self) -> Optional['ProcessNode']:
        """
        Return the process node that called this process node, or None if it does not have a caller

        :returns: process node that called this process node instance or None
        """
        try:
            caller = self.get_incoming(link_type=(LinkType.CALL_CALC, LinkType.CALL_WORK)).one().node
        except ValueError:
            return None
        else:
            return caller

    def validate_incoming(self, source: Node, link_type: LinkType, link_label: str) -> None:
        """Validate adding a link of the given type from a given node to ourself.

        Adding an input link to a `ProcessNode` once it is stored is illegal because this should be taken care of
        by the engine in one go. If a link is being added after the node is stored, it is most likely not by the engine
        and it should not be allowed.

        :param source: the node from which the link is coming
        :param link_type: the link type
        :param link_label: the link label
        :raise TypeError: if `source` is not a Node instance or `link_type` is not a `LinkType` enum
        :raise ValueError: if the proposed link is invalid
        """
        super().validate_incoming(source, link_type, link_label)
        if self.is_stored:
            raise ValueError('attempted to add an input link after the process node was already stored.')

    @property
    def is_valid_cache(self) -> bool:
        """
        Return whether the node is valid for caching

        :returns: True if this process node is valid to be used for caching, False otherwise
        """
        if not (super().is_valid_cache and self.is_finished):
            return False

        try:
            process_class = self.process_class
        except ValueError as exc:
            self.logger.warning(f"Not considering {self} for caching, '{exc!r}' when accessing its process class.")
            return False

        # For process functions, the `process_class` does not have an is_valid_cache attribute
        try:
            is_valid_cache_func = process_class.is_valid_cache
        except AttributeError:
            return True

        return is_valid_cache_func(self)

    @is_valid_cache.setter
    def is_valid_cache(self, valid: bool) -> None:
        """Set whether this node instance is considered valid for caching or not.

        :param valid: whether the node is valid or invalid for use in caching.
        """
        super().is_valid_cache = valid  # type: ignore[misc]

    def _get_objects_to_hash(self) -> List[Any]:
        """
        Return a list of objects which should be included in the hash.
        """
        res = super()._get_objects_to_hash()
        res.append({
            entry.link_label: entry.node.get_hash()
            for entry in self.get_incoming(link_type=(LinkType.INPUT_CALC, LinkType.INPUT_WORK))
            if entry.link_label not in self._hash_ignored_inputs
        })
        return res
